import pytest
import random

pytestmark = [pytest.mark.web_api]


@pytest.mark.dist
@pytest.mark.asyncio
async def test_tcp_cluster_details(dist_ctx, base_url, http_client, default_token):

    worker_set = dist_ctx.executor.get_available_workers()
    host1, host2 = sorted(worker_set.hosts())
    url = f"{base_url}/api/config/connection/?token={default_token}"

    conn_details = {
        'connection': {
            'type': 'TCP',
            'address': 'tcp://scheduler:8786',
        }
    }
    async with http_client.put(url, json=conn_details) as response:
        assert response.status == 200

    url = f"{base_url}/api/config/cluster/?token={default_token}"
    async with http_client.get(url) as response:
        assert response.status == 200
        details = await response.json()
        assert details == {
            "status": "ok",
            "messageType": "CLUSTER_DETAILS",
            "details": [
                {
                    'cpu': 2,
                    'cuda': 0,
                    'host': host1,
                    'service': 1
                },
                {
                    'cpu': 2,
                    'cuda': 0,
                    'host': host2,
                    'service': 1
                }
            ]
        }


@pytest.mark.asyncio
async def test_local_cluster_details(shared_state, base_url, http_client, default_token):
    num_workers = random.randint(1, 4)
    url = f"{base_url}/api/config/connection/?token={default_token}"
    conn_details = {
        'connection': {
            'type': 'local',
            'numWorkers': num_workers,
        }
    }

    async with http_client.put(url, json=conn_details) as response:
        assert response.status == 200

    executor = await shared_state.executor_state.get_executor()
    worker_set = await executor.get_available_workers()
    host = worker_set.hosts().pop()

    url = f"{base_url}/api/config/cluster/?token={default_token}"
    async with http_client.get(url) as response:
        assert response.status == 200
        details = await response.json()
        assert details == {
            "status": "ok",
            "messageType": "CLUSTER_DETAILS",
            "details": [{
                "host": host,
                "cpu": num_workers,
                "cuda": 0,
                "service": 1
            }]
        }



import logging
from typing import TYPE_CHECKING, Callable, TypeVar, Union

import uuid
import typing_extensions

from libertem.analysis.base import AnalysisResultSet

from .models import AnalysisInfo, AnalysisResultInfo, CompoundAnalysisInfo

if TYPE_CHECKING:
    from .state import SharedState
    from .engine import JobEngine


log = logging.getLogger(__name__)

T = TypeVar('T')


class RPCContext:
    """
    A high-level interface of methods available to RPC procedures.

    An RPCContext is always directly connected to a specific compound analysis,
    but it can access other analyses, too.

    Under the hood, this connects the RPC procedure to the current
    server-side state of the web API, and allows access to information
    about analyses, and allows to run an analysis.
    """
    def __init__(
        self, state: "SharedState", compound_analysis_id: str, engine: "JobEngine"
    ) -> None:
        self.state = state
        self.compound_analysis_id = compound_analysis_id
        self.engine = engine

    def get_compound_analysis(self) -> CompoundAnalysisInfo:
        """
        Get information about the compound analysis in question.

        Returns
        -------
        CompoundAnalysisInfo
            Information about the analysis which this RPC is in connection with.
        """
        return self.state.compound_analysis_state[self.compound_analysis_id]

    def have_analysis_results(self, analysis_id: str) -> bool:
        return self.state.analysis_state.have_results(analysis_id)

    def get_analysis_details(self, analysis_id: str) -> AnalysisInfo:
        return self.state.analysis_state[analysis_id]

    def get_analysis_results(
        self, analysis_id: str
    ) -> AnalysisResultInfo:
        results = self.state.analysis_state.get_results(analysis_id)
        return results

    async def run_analysis(self, analysis_id: str) -> AnalysisResultSet:
        job_id = str(uuid.uuid4())
        return await self.engine.run_analysis(analysis_id, job_id)

    async def run_sync(self, fn: Callable[..., T], *args, **kwargs) -> T:
        """
        To run a blocking, more compute-intensive function as part of
        a RPC, please wrap the call with this function!
        """
        return await self.engine.run_sync(fn, *args, **kwargs)


class SyncProcedureProtocol(typing_extensions.Protocol):
    def __call__(self, rpc_context: RPCContext):
        """
        The server-side interface that each procedure needs to implement

        Parameters
        ----------
        rpc_context : RPCContext
            Information about the related compound analysis is contained in this
            instance of :code:`RPCContext`, including methods to run analyses.
        """
        ...


class AsyncProcedureProtocol(typing_extensions.Protocol):
    async def __call__(self, rpc_context: RPCContext):
        """
        The server-side interface that each procedure needs to implement (async version)

        Parameters
        ----------
        rpc_context : RPCContext
            Information about the related compound analysis is contained in this
            instance of :code:`RPCContext`, including methods to run analyses.
        """
        ...


ProcedureProtocol = Union[AsyncProcedureProtocol, SyncProcedureProtocol]



import threading
from typing import TYPE_CHECKING, Callable, Any, NamedTuple, Optional
from collections.abc import Iterable
import time

from libertem.common.executor import WorkerQueueEmpty


if TYPE_CHECKING:
    from libertem.common.executor import WorkerQueue, TaskCommHandler
    from libertem.udf.base import UDFTask
    from libertem.io.dataset.base.tiling import DataTile
    from libertem.io.dataset.base.partition import Partition


class CommsDispatcher:
    """
    Monitors a :code:`WorkerQueue` in a background thread
    and launches callbacks in response to messages recieved
    Callbacks are registered as a dictionary of subscriptions
        {topic_name: [callback, ...]]}
    and are called in order they were recieved in the same thread
    that is doing the monitoring. The callbacks should be lightweight
    in order to not build up too many messages in the queue

    The functionality of this class mirrors Dask's structured logs
    feature, which has a similar message => topic => callback
    model running in the client event loop
    """
    def __init__(self, queue: 'WorkerQueue', subscriptions: dict[str, list[Callable]]):
        self._message_q = queue
        self._subscriptions = subscriptions
        self._thread = None

    def __enter__(self, *args, **kwargs):
        if self._thread is not None:
            raise RuntimeError('Cannot re-enter CommsDispatcher')
        self._thread = threading.Thread(
            target=self.monitor_queue,
            name="CommsDispatcher",
        )
        self._thread.daemon = True
        self._thread.start()

    def __exit__(self, *args, **kwargs):
        if self._thread is None:
            return
        self._message_q.put(('STOP', {}))
        self._thread.join()
        self._thread = None
        # Drain queue just in case
        while True:
            try:
                with self._message_q.get(block=False) as _:
                    ...
            except WorkerQueueEmpty:
                break

    def monitor_queue(self):
        """
        Monitor the queue for messages
        If there are no subscribers this should drain
        messages from the queue as fast as they are recieved
        """
        while True:
            with self._message_q.get(block=True) as ((topic, msg), _):
                if topic == 'STOP':
                    break
                try:
                    for callback in self._subscriptions[topic]:
                        callback(topic, msg)
                except KeyError:
                    pass


class ProgressState(NamedTuple):
    """
    Container for progress state, used to communicate
    from ProgressManager to ProgressReporter
    """
    #: float: Number of frames processed
    num_frames_complete: float
    #: int: Total number of frames to process
    num_frames_total: int
    #: int: Number of partitions completed
    num_part_complete: int
    #: int: Number of partitions in-progress
    num_part_in_progress: int
    #: int: Total number of partitions
    num_part_total: int
    #: str: A unique string identifier for the job associated
    #: with this progress message
    progress_id: str


class ProgressReporter:
    """
    Interface for progress bar display / updating

    This class will receive :class:`ProgressState`
    instances to notify it about the start, progression
    and end of a job submitted to the :code:`UDFRunner`.
    The implementation should be adapted to display or
    log the progress as required for the use case.

    It is possible that multiple jobs are submitted
    to a single executor at the same time and therefore
    the implementation should ensure that concurrent instances
    of the class display correctly, or that the same instance
    of the class can handle updates from multiple threads
    concurrently. Each :class:`ProgressState` message contains a field
    :code:`progress_id` which is unique to each job, and therefore
    the implementation can use this to distinguish updates from
    multiple sources.
    """
    def __init__(self):
        raise NotImplementedError()

    def start(self, state: ProgressState):
        """
        Signal the creation of a new job with the expected
        number of partitions and frames, and unique progress_id string.
        """
        raise NotImplementedError()

    def update(self, state: ProgressState):
        """
        Signal an intermediate update to the progress of a job
        """
        raise NotImplementedError()

    def end(self, state: ProgressState):
        """
        Signal the end of a given job

        This method will always be called and any updates
        recieved after this message should be ignored.
        """
        raise NotImplementedError()


class TQDMProgressReporter(ProgressReporter):
    """
    Progress bar display via tqdm

    Supports concurrent usage of multiple instances of
    this class, but does not handle multi-threaded use
    of the same instance to report multiple jobs.
    """
    def __init__(self):
        self._bar = None
        # Integers used to check if bar description should be changed
        # integers because faster than strcmp, updated in _should_update_description
        self._desc_key = (-1, -1, -1)

    def start(self, state: ProgressState):
        from tqdm.auto import tqdm
        self._bar = tqdm(desc=self._get_description(state),
                         total=state.num_frames_total,
                         leave=True)

    def update(self, state: ProgressState):
        return self._update(state, clip=True, refresh=False)

    def _update(self, state: ProgressState, *, clip: bool, refresh: bool):
        if state.num_frames_total != self._bar.total:
            # Should never happen but handle just in case
            self._bar.total = state.num_frames_total
            self._bar.refresh()
        increment = self._get_increment(state, clip=clip)
        if increment > 0:
            self._bar.update(increment)
        if self._should_update_description(state):
            self._bar.set_description(self._get_description(state))
        if refresh:
            self._bar.refresh()

    def end(self, state: ProgressState):
        self._update(state, clip=True, refresh=True)
        self._bar.close()

    def _should_update_description(self, state: ProgressState) -> bool:
        """
        Check the state to see if the elements used by self._get_description
        have changed, and if so update our record (self._desc_key) and return True
        """
        new_desc_key = (
            state.num_part_complete,
            state.num_part_in_progress,
            state.num_part_total
        )
        should_update = new_desc_key != self._desc_key
        if should_update:
            self._desc_key = new_desc_key
        return should_update

    @staticmethod
    def _get_description(state: ProgressState) -> str:
        """
        Get the most recent description string for the
        bar, including partition information
        If we know that partitions are in progress
        include this in parentheses after n_completed
        """

        if state.num_part_in_progress:
            return (f'Partitions {state.num_part_complete}({state.num_part_in_progress})'
                    f'/{state.num_part_total}, Frames')
        else:
            return (f'Partitions {state.num_part_complete}'
                    f'/{state.num_part_total}, Frames')

    def _get_increment(self, state: ProgressState, clip: bool = True):
        """
        Get the increment to apply to the progress bar based on current state
        and the state of the bar itself (bar.n is the total as-tracked by tqdm)

        Assumes self._bar.total has first been updated
        to state.num_frames_total if this is necessary
        """
        increment = int(state.num_frames_complete) - self._bar.n
        if clip:
            max_update = self._bar.total - self._bar.n
        else:
            max_update = increment + 1
        return max(0, min(increment, max_update))


class ProgressManager:
    """
    Handle updating of a progress reporter for a set of :code:`UDFTasks`, to be
    completed in any order. By default constructs a
    :code:`TQDMProgressReporter`, if no instance is passed in.

    The bar displays as such:

        Partitions: n_complete(n_in_progress) / n_total, ...\
            Frames: [XXXXX..] frames_completed / total_frames ...

    When processing tile stacks, stacks are treated as frames
    as such: (pseudo_frames = tile.size // sig_size)

    The bar will render in a Jupyter notebook as a JS widget
    automatically via tqdm.auto
    """
    def __init__(
        self,
        tasks: Iterable['UDFTask'],
        progress_id: str,
        reporter: Optional[ProgressReporter] = None,
    ):
        if not tasks:
            raise ValueError('Cannot display progress for empty tasks')
        self._progress_id = progress_id
        # the number of whole frames we expect each task to process
        self._task_max = {t.partition.get_ident(): t.task_frames
                          for t in tasks}
        # _counters is our record of progress on a task,
        # values are floating whole frames processed
        # as in tile mode we can process part of a frame
        self._counters = {k: 0. for k in self._task_max.keys()}
        self._total_frames = sum(self._task_max.values())
        # For converting tiles to pseudo-frames
        self._sig_size = tasks[0].partition.shape.sig.size
        # Counters for part display
        self._complete = set()
        self._in_progress = set()
        self._num_total = len(self._counters)
        if reporter is None:
            reporter = TQDMProgressReporter()
        elif not isinstance(reporter, ProgressReporter):
            # If not a ProgressReporter instance,
            # instantiate as if it has a bare __init__
            # Useful to be able to inject an instance
            # of ProgressReporter in case we need to setup
            # or access the reporter somehow (e.g. for testing)
            reporter = reporter()
        assert isinstance(reporter, ProgressReporter)
        self.reporter = reporter
        reporter.start(self.state)

    @property
    def state(self) -> ProgressState:
        return ProgressState(
            sum(self._counters.values()),
            self._total_frames,
            len(self._complete),
            len(self._in_progress),
            self._num_total,
            self._progress_id,
        )

    def finalize_task(self, task: 'UDFTask'):
        """
        When a task completes and we recieve its results on
        the main node, this is called to update the partition
        progress counters and frame counter in case we didn't
        recieve a complete history of the partition yet
        """
        topic = 'partition_complete'
        ident = task.partition.get_ident()
        message = {'ident': task.partition.get_ident()}
        if ident in self._task_max:
            self.handle_end_task(topic, message)

    def close(self):
        self.reporter.end(self.state)

    def connect(self, comms: 'TaskCommHandler'):
        """
        Register the callbacks on this class with the TaskCommHandler
        which will be dispatching messages recieved from the tasks
        """
        comms.subscribe('partition_start', self.handle_start_task)
        comms.subscribe('partition_complete', self.handle_end_task)
        comms.subscribe('tile_complete', self.handle_tile_update)

    def handle_start_task(self, topic: str, message: dict[str, Any]):
        """
        Increment the num_in_progress counter

        # NOTE An extension to this would be to track
        the identities of partitions in progress / completed
        for a richer display / more accurate accounting
        """
        if topic != 'partition_start':
            raise RuntimeError('Unrecognized topic')
        t_id = message['ident']
        if t_id not in self._complete:
            # if not complete handles case task was completed
            # before we can process its start message,
            # completion is signalled in the main thread
            # while start messages are processed in the background
            self._in_progress.add(t_id)
        self.reporter.update(self.state)

    def handle_end_task(self, topic: str, message: dict[str, Any]):
        """
        Increment the counter for the task to the max value
        and update the various counters / description
        """
        if topic != 'partition_complete':
            raise RuntimeError('Unrecognized topic')
        t_id = message['ident']
        remain = self._task_max[t_id] - int(self._counters[t_id])
        if remain:
            self._counters[t_id] = self._task_max[t_id]
        self._in_progress.discard(t_id)
        self._complete.add(t_id)
        self.reporter.update(self.state)

    def handle_tile_update(self, topic: str, message: dict[str, Any]):
        """
        Update the frame progress counter for the task
        and push the increment to the progress reporter

        Tile stacks are converted to pseudo-frames via the sig_size
        """
        if topic != 'tile_complete':
            raise RuntimeError('Unrecognized topic')
        t_id = message['ident']
        if self._counters[t_id] >= self._task_max[t_id]:
            return
        elements = message['elements']
        pframes = elements / self._sig_size
        self._counters[t_id] += pframes
        self.reporter.update(self.state)


class PartitionTrackerNoOp:
    """
    A no-op class matching the PartitionProgressTracker interface
    Used when progress == False to avoid any additional overhead
    """
    def signal_start(self, *args, **kwargs):
        ...

    def signal_tile_complete(self, *args, **kwargs):
        ...

    def signal_complete(self, *args, **kwargs):
        ...


def get_time():
    # Exists for testing / mocking
    return time.time()


class PartitionProgressTracker(PartitionTrackerNoOp):
    """
    Tracks the tile processing speed of a Partition and
    dispatches messages via the worker_context.signal() method
    under certain conditions

    Parameters
    ----------
    partition : Partition
        The partition to track progress for
    min_message_interval : float, optional
        The minumum time between messages, by default 1 second.
    """
    def __init__(
        self,
        partition: 'Partition',
        min_message_interval: float = 1.,
    ):
        self._ident = partition.get_ident()
        try:
            self._worker_context = partition._worker_context
        except AttributeError:
            self._worker_context = None

        # Counters to track / rate-limit messages
        self._elements_complete = 0
        self._last_message_t = None
        self._min_message_interval = min_message_interval

    def signal_start(self):
        """
        Signal that the partition has begun processing
        """
        if self._worker_context is None:
            return
        self._worker_context.signal(
            self._ident,
            'partition_start',
            {},
        )

    def signal_tile_complete(self, tile: 'DataTile'):
        """
        Register that tile.size more elements have been processed
        and if certain condition are met, send a signal
        """
        if self._worker_context is None:
            return

        send_elements = self.should_send_progress(tile.size)
        if send_elements:
            self._worker_context.signal(
                self._ident,
                'tile_complete',
                {'elements': send_elements},
            )

    def signal_complete(self):
        """
        Signal that the partition has completed processing

        This is not currently called as partition completion
        is registered on the main node as a fallback
        """
        if self._worker_context is None:
            return
        self._worker_context.signal(
            self._ident,
            'partition_complete',
            {},
        )

    def should_send_progress(self, elements: int) -> int:
        """
        Given the number elements of data that have been processed since
        the last message was sent, decide if a signal should be sent to the
        main node about the partition progress
        """
        current_t = get_time()
        self._elements_complete += elements

        if self._last_message_t is None:
            # Never send a message for the first tile stack
            # as this might have warmup overheads associated
            # Include the first elements in the history, however,
            # to give a better accounting. The first tile stack
            # is essentially treated as 'free'.
            self._last_message_t = current_t
            return 0

        time_since_last_m = current_t - self._last_message_t
        not_rate_limited = time_since_last_m > self._min_message_interval

        if not_rate_limited:
            completed_elements = self._elements_complete
            self._elements_complete = 0
            self._last_message_t = current_t
            return completed_elements

        return 0



import math
import logging
import warnings
from typing import TYPE_CHECKING, Optional, Union
from collections.abc import Sequence
from typing_extensions import Literal

import numpy as np

from libertem.common.exceptions import UDFException
from libertem.io.corrections import CorrectionSet
from libertem.common import Shape, Slice
from libertem.common.math import prod
from libertem.common.udf import UDFProtocol, UDFMethod

if TYPE_CHECKING:
    from numpy import typing as nt
    from libertem.io.dataset.base import DataSet, Partition

log = logging.getLogger(__name__)

TilingIntent = Union[Literal["partition"], Literal["frame"], Literal["tile"]]


class TilingScheme:
    def __init__(
        self, slices: list[Slice],
        tileshape: Shape, dataset_shape: Shape, intent: Optional[TilingIntent] = None, debug=None
    ):
        self._slices = slices
        self._tileshape = tileshape
        self._dataset_shape = dataset_shape
        self._debug = debug
        self._intent = intent

        if tileshape.nav.dims > 1:
            raise ValueError("tileshape should have flat navigation dimensions")

    def adjust_for_partition(self, partition: "Partition") -> "TilingScheme":
        """
        If the intent is per-partition processing, the tiling scheme must match the
        partition shape exactly. If there is a mismatch, this method returns a
        new scheme that matches the partition.

        Parameters
        ----------
        partition
            The Partition we want to adjust the tiling scheme to.

        Returns
        -------
        TilingScheme
            The adjusted tiling scheme, or this one, if it matches exactly
        """
        partition_size = partition.slice.shape.nav.size
        if partition_size != self.depth and self.intent == "partition":
            # adjust depth to match partition size exactly:
            new_shape = Shape(
                (partition_size,) + tuple(self._tileshape.sig),
                sig_dims=self._tileshape.sig.dims
            )
            return TilingScheme(
                slices=self._slices,
                tileshape=new_shape,
                dataset_shape=self._dataset_shape,
                intent=self._intent,
                debug=self._debug,
            )
        return self

    @classmethod
    def make_for_shape(
        cls,
        tileshape: Shape,
        dataset_shape: Shape,
        intent: Optional[TilingIntent] = None,
        debug=None,
    ) -> "TilingScheme":
        """
        Make a TilingScheme from `tileshape` and `dataset_shape`.

        Note that both in signal and navigation direction there are border
        effects, i.e. if the depth doesn't evenly divide the number of
        frames in the partition (simplified, ROI also applies...), or if
        the signal dimensions of `tileshape` doesn't evenly divide the signal
        dimensions of the `dataset_shape`.

        Parameters
        ----------
        tileshape
            Uniform shape of all tiles. Should have flat navigation axis
            (meaning tileshape.nav.dims == 1) and be contiguous in signal
            dimensions.

        dataset_shape
            Shape of the whole data set. Only the signal part is used.

        intent
            The intent of this scheme (whole partitions, frames or tiles)
            Needs to be set for correct per-partition tiling!
        """
        # FIXME: validate navigation part of the tileshape to be contiguous
        # (i.e. a shape like (1, 1, ..., 1, X1, ..., XN))
        # where X1 is <= the dataset shape at that index, and X2, ..., XN are
        # equal to the dataset shape at that index

        sig_slice = Slice(
            origin=tuple([0] * dataset_shape.sig.dims),
            shape=dataset_shape.sig
        )
        subslices = list(sig_slice.subslices(tileshape.sig))
        return cls(
            slices=subslices,
            tileshape=tileshape,
            dataset_shape=dataset_shape,
            debug=debug,
            intent=intent,
        )

    def __getitem__(self, idx: int) -> Slice:
        return self._slices[idx]

    def __len__(self):
        return len(self._slices)

    def __repr__(self):
        unique_shapes = list({tuple(slice_.shape) for slice_ in self._slices})
        return "<TilingScheme (depth=%d) shapes=%r len=%d>" % (
            self.depth, unique_shapes, len(self._slices),
        )

    @property
    def intent(self) -> Optional[TilingIntent]:
        return self._intent

    @property
    def slices(self):
        """
        signal-only slices for all possible positions
        """
        return list(enumerate(self._slices))

    @property
    def slices_array(self):
        """
        Returns the slices from the schema as a numpy ndarray
        `a` of shape `(n, 2, sig_dims)` with:
        `a[i, 0]` are origins for slice `i`
        `a[i, 1]` are shapes for slice `i`
        """
        sig_dims = self._tileshape.sig.dims
        slices = np.zeros((len(self), 2, sig_dims), dtype=np.int64)
        for idx, slice_ in self.slices:
            slices[idx] = (tuple(slice_.origin), tuple(slice_.shape))
        return slices

    @property
    def shape(self):
        """
        tileshape. note that some border tiles can be smaller!
        """
        return self._tileshape

    @property
    def dataset_shape(self):
        return self._dataset_shape

    @property
    def depth(self):
        return self._tileshape.nav[0]


class Negotiator:
    """
    Tile shape negotiator. The main functionality is in `get_scheme`,
    which, given a `udf`, `dataset` and `read_dtype` will generate
    a `TilingScheme` that is compatible with both the `UDF` and the
    `DataSet`, possibly even optimal.
    """

    def validate(
        self,
        shape: tuple[int, ...],
        ds_sig_shape: tuple[int, ...],
        size: int,
        io_max_size: int,
        itemsize: int,
        base_shape: tuple[int, ...],
        corrections: Optional[CorrectionSet],
    ):
        sig_shape = shape[1:]
        # we need some wiggle room with the size, because there may be a harder
        # lower size value for some cases (for example HDF5, which overrides
        # some of the sizing negotiation we are doing here)
        if any(s > ps for s, ps in zip(sig_shape, ds_sig_shape)):
            raise ValueError("generated tileshape does not fit the partition")
        size_px = max(size, io_max_size) // itemsize
        if prod(shape) > size_px:
            message = "shape %r (%d) does not fit into size %d" % (
                shape, prod(shape), size_px
            )
            # The shape might be exceeded if dead pixel correction didn't find a
            # valid tiling scheme. In that case it falls back to by-frame processing.
            if (
                corrections is not None
                and corrections.get_excluded_pixels() is not None
                and shape[0] == 1
            ):
                warnings.warn(message)
            else:
                raise ValueError(message)
        for dim in range(len(base_shape)):
            # Tile shape always has one nav dim
            # Allow only base shape for nav
            # Allow full frames for sig
            if ((shape[dim] % base_shape[dim] != 0)
                    and (dim == 0 or (dim > 0 and shape[dim] != ds_sig_shape[dim - 1]))):
                raise ValueError(
                    f"The tileshape {shape} is incompatible with base "
                    f"shape {base_shape} and dataset shape {ds_sig_shape} in dimension {dim}."
                )

    def get_scheme(
            self,
            udfs: Sequence[UDFProtocol],
            dataset,
            read_dtype: "nt.DTypeLike",
            approx_partition_shape: Shape,
            roi: Optional[np.ndarray] = None,
            corrections: Optional[CorrectionSet] = None,
    ) -> TilingScheme:
        """
        Generate a :class:`TilingScheme` instance that is
        compatible with both the given `udf` and the
        :class:~`libertem.io.dataset.base.DataSet`.

        Parameters
        ----------

        udfs : Sequence[UDFProtocol]
            The concrete UDFs to optimize the tiling scheme for.
            Depending on the method (tile, frame, partition)
            and preferred total input size and depth.

        dataset : DataSet
            The DataSet instance we generate the scheme for.

        read_dtype
            The dtype in which the data will be fed into the UDF

        approx_partition_shape
            The approximate partition shape that is likely to be used

        roi : np.ndarray
            Region of interest

        corrections : CorrectionSet
            Correction set to consider in negotiation
        """
        itemsize = np.dtype(read_dtype).itemsize

        # FIXME: let the UDF define upper bound for signal size (lower bound, too?)
        # (signal buffers should fit into the L2 cache)
        # try not to waste page faults:
        min_sig_size = dataset.get_min_sig_size()
        ds_sig_shape = dataset.shape.sig

        # This already takes corrections into account through a different pathway
        need_decode = dataset.need_decode(roi=roi, read_dtype=read_dtype, corrections=corrections)

        io_max_size = self._get_io_max_size(dataset, approx_partition_shape, itemsize, need_decode)

        depths = [
            self._get_min_depth(udf, approx_partition_shape)
            for udf in udfs
        ]
        depth = max(depths)  # take the largest min-depth
        base_shape = self._get_base_shape(udfs, dataset, approx_partition_shape, roi)

        intent = self._get_intent(udfs)

        sizes = [
            self._get_size(
                io_max_size, udf, itemsize, approx_partition_shape, base_shape,
            )
            for udf in udfs
        ]
        if intent == "partition":
            size = max(sizes)  # by partition wants to be big, ...
        else:
            size = min(sizes)
        size_px = size // itemsize

        if corrections is not None and corrections.have_corrections():
            # The correction has to make sure that there are no excluded pixels
            # at tile boundaries
            base_shape = corrections.adjust_tileshape(
                tile_shape=base_shape,
                sig_shape=tuple(ds_sig_shape),
                base_shape=base_shape,
            )

        # first, scale `base_shape` up to contain at least `min_sig_size` items:
        min_factors = self._get_scale_factors(
            base_shape,
            containing_shape=ds_sig_shape,
            size=min_sig_size,
        )

        min_base_shape = self._scale_base_shape(base_shape, min_factors)

        # considering the min size, calculate the max depth:
        max_depth = max(1, size_px // prod(min_base_shape))
        if depth > max_depth:
            depth = max_depth

        full_base_shape = (1,) + tuple(base_shape)
        min_factors = (depth,) + tuple(min_factors)

        containing_shape = approx_partition_shape

        factors = self._get_scale_factors(
            full_base_shape,
            containing_shape=containing_shape,
            size=size_px,
            min_factors=min_factors,
        )
        tileshape = self._scale_base_shape(full_base_shape, factors)
        tileshape_orig = tileshape

        # the dataset has a "veto" on the tileshape:
        # FIXME: this veto may break if the base shape was adjusted
        # above, and we need to be careful not to break corrections with this,
        # and also fulfill requests of per-frame reading
        log.debug("tileshape before adjustment: %r", (tileshape,))
        tileshape = tuple(dataset.adjust_tileshape(tileshape, roi=roi))
        log.debug("tileshape after adjustment: %r", (tileshape,))

        # if the veto generated a tileshape that is smaller than the full base shape,
        # we need to re-adjust the full_base_shape
        if tileshape_orig != tileshape:  # make sure we don't change too eagerly:
            if tileshape[0] < full_base_shape[0]:
                full_base_shape = (tileshape[0], *full_base_shape[1:])
            has_pixel_corr = (
                corrections is not None
                and corrections.get_excluded_pixels() is not None
            )
            for (orig, new, sig) in zip(tileshape_orig[1:], tileshape[1:], ds_sig_shape):
                if new != orig and new != sig:
                    # Otherwise we may generate incorrect correction results
                    err_str = (
                        "dataset.adjust_tileshape() can only accept tile sig shape or switch to"
                        "full frames if dead pixel patching is active. "
                        f"Got original tile shape {tileshape_orig}, new tileshape {tileshape} "
                        f"and dataset sig shape {ds_sig_shape}"
                    )
                    if has_pixel_corr:
                        raise ValueError(err_str)
                    else:
                        warnings.warn(err_str)

        self.validate(
            tileshape, ds_sig_shape, size, io_max_size, itemsize, full_base_shape, corrections,
        )
        return TilingScheme.make_for_shape(
            tileshape=Shape(tileshape, sig_dims=ds_sig_shape.dims),
            dataset_shape=dataset.shape,
            intent=intent,
            debug={
                "min_factors": min_factors,
                "factors": factors,
                "tileshape": tileshape,
                "size": size,
                "size_px": size_px,
                "full_base_shape": full_base_shape,
                "need_decode": need_decode,
                "depth": depth,
            }
        )

    def _get_io_max_size(self, dataset, approx_partition_shape, itemsize, need_decode):
        if need_decode:
            io_max_size = dataset.get_max_io_size()
            if io_max_size is None:
                io_max_size = 2**20
        else:
            io_max_size = itemsize * prod(approx_partition_shape)
        return io_max_size

    def _get_scale_factors(self, shape, containing_shape, size, min_factors=None):
        """
        Generate scaling factors to scale `shape` up to `size` elements,
        while being constrained to `containing_shape`.
        """
        log.debug(
            "_get_scale_factors in: shape=%r, containing_shape=%r, size=%r, min_factors=%r",
            shape, containing_shape, size, min_factors
        )
        assert len(shape) == len(containing_shape)
        if min_factors is None:
            factors = [1] * len(shape)
        else:
            factors = list(min_factors)
        max_factors = tuple(
            cs // s
            for s, cs in zip(shape, containing_shape)
        )
        prelim_shape = self._scale_base_shape(shape, factors)
        rest = size / prod(prelim_shape)
        if rest < 1:
            rest = 1
        for idx in range(len(shape)):
            max_factor = max_factors[idx]
            factor = int(math.floor(rest * factors[idx]))
            if factor < factors[idx]:
                factor = factors[idx]
            if factor > max_factor:
                factor = max_factor
            factors[idx] = factor
            prelim_shape = self._scale_base_shape(shape, factors)
            rest = max(1, math.floor(size / prod(prelim_shape)))
        log.debug(
            "_get_scale_factors out: %r",
            factors,
        )
        return factors

    def _scale_base_shape(self, base_shape, factors):
        assert len(factors) == len(base_shape)
        return tuple(
            f * bs
            for f, bs in zip(factors, base_shape)
        )

    def _get_default_size(self):
        # FIXME: adjust size to L3 // number of workers per node
        return 1*2**20

    def _get_udf_size_pref(self, udf: UDFProtocol):
        udf_prefs = udf.get_tiling_preferences()
        size = udf_prefs.get("total_size", np.inf)
        if size is UDFProtocol.TILE_SIZE_BEST_FIT:
            size = self._get_default_size()
        return size

    def _get_intent(self, udfs: Sequence[UDFProtocol]) -> TilingIntent:
        udf_methods = tuple(udf.get_method() for udf in udfs)
        if any(m not in tuple(UDFMethod) for m in udf_methods):
            raise UDFException('A UDF declared an invalid processing method')
        if UDFMethod.PARTITION in udf_methods:
            return "partition"
        elif UDFMethod.FRAME in udf_methods:
            return "frame"
        elif UDFMethod.TILE in udf_methods:
            return "tile"
        else:
            raise ValueError('No recognized UDF method, empty udfs arg?')

    def _get_size(
            self, io_max_size, udf: UDFProtocol, itemsize,
            approx_partition_shape: Shape, base_shape):
        """
        Calculate the maximum tile size in bytes
        """
        udf_method = udf.get_method()
        partition_size = itemsize * prod(tuple(approx_partition_shape))
        partition_size_sig = itemsize * prod(tuple(approx_partition_shape.sig))
        if udf_method == UDFMethod.FRAME:
            size = max(self._get_default_size(), partition_size_sig)
        elif udf_method == UDFMethod.PARTITION:
            size = partition_size
        elif udf_method == UDFMethod.TILE:
            # start with the UDF size preference:
            size = self._get_udf_size_pref(udf)

            # constrain to maximum read size
            size = min(size, io_max_size)

            # if the base_shape is larger than the current maximum size,
            # we need to increase the size:
            base_size = itemsize * prod(base_shape)
            size = max(base_size, size)
        else:  # pragma: no cover
            # Should never be reached, this is checked earlier in UDFRunner
            raise UDFException(f'UDF.get_method() returned unrecognized method: {udf_method}')
        return size

    def _get_base_shape(
        self,
        udfs: Sequence["UDFProtocol"],
        dataset: "DataSet",
        approx_partition_shape: Shape,
        roi: Optional[np.ndarray],
    ):
        methods = [
            udf.get_method()
            for udf in udfs
        ]
        if any(m in (UDFMethod.FRAME, UDFMethod.PARTITION) for m in methods):
            base_shape = approx_partition_shape.sig
        else:
            # only by tile:
            base_shape = Shape(
                dataset.get_base_shape(roi=roi),
                sig_dims=approx_partition_shape.sig.dims
            ).sig
        return base_shape

    def _get_udf_depth_pref(self, udf: "UDFProtocol", approx_partition_shape: Shape) -> int:
        udf_prefs = udf.get_tiling_preferences()
        depth = udf_prefs.get("depth", UDFProtocol.TILE_DEPTH_DEFAULT)
        if depth is UDFProtocol.TILE_DEPTH_DEFAULT:
            depth = 32
        if depth > approx_partition_shape[0]:
            depth = approx_partition_shape[0]
        return depth

    def _get_min_depth(self, udf: "UDFProtocol", approx_partition_shape: Shape) -> int:
        udf_method = udf.get_method()

        if udf_method == UDFMethod.PARTITION:
            return approx_partition_shape[0]
        elif udf_method == UDFMethod.TILE:
            return self._get_udf_depth_pref(udf, approx_partition_shape)
        return 1



import logging
import socket

import click

log_values = "Allowed values are 'critical', 'error', 'warning', 'info', 'debug'."


preload_help = (
    'Module, file or code to preload on workers, for example HDF5 plugins. '
    'Can be specified multiple times. See also '
    'https://docs.dask.org/en/stable/how-to/customize-initialization.html#preload-scripts '
    'for the behavior with Dask workers (current default)'
    'and https://libertem.github.io/LiberTEM/reference/dataset.html#hdf5 '
    'for information on loading HDF5 files that depend on custom filters.'
)


@click.command()
@click.argument('scheduler', default="tcp://localhost:8786", type=str)
@click.option('-k', '--kind', help='Worker kind. Currently only "dask" is implemented.',
              default="dask", type=str)
@click.option('-d', '--local-directory', help='local directory to manage temporary files',
              default='dask-worker-space', type=str)
@click.option('-c', '--n-cpus', type=int, default=None,
              help='Number of CPUs to use, defaults to number of CPU cores without hyperthreading.')
@click.option('-u', '--cudas', type=str, default=None,
              help='List of CUDA device IDs to use, defaults to all detected CUDA devices. '
              'Use "" to deactivate CUDA.')
@click.option('-p', '--has-cupy', type=bool, default=None,
              help='Activate CuPy integration, defaults to detection of installed CuPy module.')
@click.option('-n', '--name', help='Name of the cluster node, defaults to host name',
              type=str)
@click.option('-l', '--log-level', help=f"set logging level. Default is 'info'. {log_values}",
              default='INFO')
@click.option('--preload', help=preload_help,
              default=None, type=str, multiple=True)
def main(kind, scheduler, local_directory, n_cpus, cudas,
         has_cupy, name, log_level, preload: tuple[str, ...]):
    from libertem.common.threading import set_num_threads_env
    with set_num_threads_env(1):
        from libertem.utils.devices import detect
        from libertem.cli_tweaks import console_tweaks
        from libertem.executor.dask import cli_worker
        console_tweaks()

        if kind != 'dask':
            raise NotImplementedError(
                f"Currently only worker kind 'dask' is implemented, got {kind}."
            )

        numeric_level = getattr(logging, log_level.upper(), None)
        if not isinstance(numeric_level, int):
            raise click.UsageError(f'Invalid log level: {log_level}.\n{log_values}')

        defaults = detect()

        if n_cpus is None:
            cpus = list(defaults['cpus'])
        else:
            cpus = list(range(n_cpus))

        if cudas == '':
            cudas = []
        elif cudas is None:
            cudas = list(defaults['cudas'])
        else:
            cudas = list(map(int, cudas.split(',')))

        if has_cupy is None:
            has_cupy = defaults['has_cupy']

        if not name:
            name = socket.gethostname()

        cli_worker(
            scheduler=scheduler,
            local_directory=local_directory,
            cpus=cpus,
            cudas=cudas,
            has_cupy=has_cupy,
            name=name,
            log_level=numeric_level,
            preload=preload,
        )



import os
import subprocess


def get_git_rev():
    # NOTE: there is a copy of this code in setup.py!
    try:
        new_cwd = os.path.abspath(os.path.dirname(__file__))
        rev_raw = subprocess.check_output(["git", "rev-parse", "HEAD"], cwd=new_cwd)
        return rev_raw.decode("utf8").strip()
    except Exception:
        return "unknown"



import logging

import numpy as np

from libertem.udf import UDF


log = logging.getLogger(__name__)


class AutoUDF(UDF):
    '''
    Generate a UDF for reduction along the signal axis from a regular function
    that processes one frame, generating result buffers automatically.

    Parameters
    ----------

    f:
        Function that accepts a frame as a single parameter. It will
        be called with np.ones(tuple(dataset.shape.sig)) to determine the output
        type and shape.
    '''
    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

    def auto_buffer(self, var):
        return self.buffer(kind='nav', extra_shape=var.shape, dtype=var.dtype)

    def get_result_buffers(self):
        '''
        Auto-generate result buffers based on the return value of f() called with a mock frame.
        '''
        mock_frame = np.ones(tuple(self.meta.dataset_shape.sig), dtype=self.meta.input_dtype)
        result = np.array(self.params.f(mock_frame))

        try:
            # FIXME Thresholds chosen somewhat arbitrarily
            if result.nbytes > max(1024, mock_frame.nbytes / (2**7)):
                log.warn(
                    "Return value of function has size %s, "
                    "not strongly reduced compared to input size %s"
                    % (result.nbytes, mock_frame.nbytes)
                )
        # Numpy arrays of dtype "object" can throw an AttributeError
        # upon size calculations
        except AttributeError:
            pass

        return {
            'result': self.auto_buffer(result)
        }

    def process_frame(self, frame):
        '''
        Call f() for the frame and assign return value to the result buffer slot.
        '''
        res = self.params.f(frame)
        self.results.result[:] = np.array(res)



import autopep8
from string import Template


class TemplateBase:
    '''
    Base class for template strings.

    Commonly used code template are available here.
    This can be override by specific code in subclasses.
    '''

    temp_ds_backend = ["io_backend = ${backend_cls}()"]
    temp_ds = ['params = $params',
               'ds = ctx.load("$type", **params)']

    temp_dep_ds = ["from libertem.io.dataset.base import ${backend_cls}"]

    temp_dep = ["import matplotlib.pyplot as plt",
                "import libertem.api as lt",
                "import numpy as np"]

    temp_dep_conn = ["import distributed as dd",
                     "from libertem.executor.dask import DaskJobExecutor"]

    temp_conn = ['client = dd.Client("$conn_url")',
                 "executor = DaskJobExecutor(client)",
                 "ctx = lt.Context(executor=executor)"]

    temp_analysis = ["${short}_analysis = ctx.$analysis_api($params)",
                     "${short}_result = ctx.run(${short}_analysis, progress=True)"]

    temp_save = ["np.save('${short}_result.npy', ${short}_result['intensity'])"]

    def code_formatter(self, code):
        return autopep8.fix_code(code)

    def format_template(self, template, data):
        template = "\n".join(template)
        return Template(template).substitute(data)



"""
Custom warning subclasses. Useful for selectively filtering. For example, to
cause an exception for all `UseDiscouragedWarning` that are thrown in the test
suite, one can run:

$ pytest -Werror:libertem.warnings.UseDiscouragedWarning tests/
"""


class UseDiscouragedWarning(FutureWarning):
    """
    This warning is thrown when a problematic feature of an API is used,
    which we currently don't have plans to remove (for backwards-compatability)
    """
    pass



from libertem.web.helpers import _convert_device_map


def test_convert_device_map():
    assert _convert_device_map({0: 3, 1: 0}) == [0, 0, 0]
    assert _convert_device_map({0: 1, 1: 1}) == [0, 1]
    assert _convert_device_map({0: 0, 1: 1, 2: 2}) == [1, 2, 2]



import numpy as np


# FIXME we might want to use an external physics or coordinate system package
# Getting this right and consistent is extremely important for correct
# interpretation of physical measurements. In particular 4D STEM data
# and ptychography can get confusing.

def make_cartesian(polar):
    '''
    Accept list of polar vectors, return list of cartesian vectors

    Parameters
    ----------
    polars : numpy.ndarray of tuples [(r1, phi1), (r2, phi2), ...]
        Polar vectors

    Returns
    -------
    numpy.ndarray of tuples [(y, x), (y, x), ...]
    '''
    xes = np.cos(polar[..., 1]) * polar[..., 0]
    yes = np.sin(polar[..., 1]) * polar[..., 0]
    return np.array((yes.T, xes.T)).T


def make_polar(cartesian):
    '''
    Accept list of cartesian vectors, return list of polar vectors

    Parameters
    ----------
    cartesian : numpy.ndarray of tuples [(y, x), (y, x), ...]
        Cartesian vectors

    Returns
    -------

    Polar vector as numpy.ndarray of tuples [(r1, phi1), (r2, phi2), ...]
    '''
    ds = np.linalg.norm(cartesian, axis=-1)
    # (y, x)
    alphas = np.arctan2(cartesian[..., 0], cartesian[..., 1])
    return np.array((ds.T, alphas.T)).T


def rotate_precalc(y, x, cos_angle, sin_angle):
    '''
    Rotate with pre-calculated entries of
    the rotation matrix. This is useful when rotating within a loop,
    in particular when combined with numba.njit of this function for inlining.

    The rotation follows https://en.wikipedia.org/wiki/Rotation_matrix

    In pixel coordinates with y pointing down and x pointing right,
    the rotation is clockwise. In physical coordinates with y pointing up instead,
    it is counter-clockwise.

    Parameters
    ==========
    y, x : float or numpy.ndarray
        Y and X components of the vector(s)

    cos_angle, sin_angle : float or numpy.ndarray
        cos(phi) and sin(phi) of rotation angle phi

    Returns
    =======
    y, x : float or numpy.ndarray
        Y and X components of rotated vector(s)

    .. versionadded:: 0.6.0
    '''
    r_x = cos_angle * x - sin_angle * y
    r_y = sin_angle * x + cos_angle * y
    return r_y, r_x


def rotate_deg(y, x, degrees):
    '''
    Rotate by angle in degrees

    The rotation follows https://en.wikipedia.org/wiki/Rotation_matrix

    In pixel coordinates with y pointing down and x pointing right,
    the rotation is clockwise. In physical coordinates with y pointing up instead,
    it is counter-clockwise.

    Parameters
    ==========
    y, x : float or numpy.ndarray
        Y and X components of the vector(s)

    degrees : float or numpy.ndarray
        Rotation angle in degrees

    Returns
    =======
    y, x : float or numpy.ndarray
        Y and X components of rotated vector(s)

    .. versionadded:: 0.6.0
    '''
    return rotate_rad(y, x, np.pi/180*degrees)


def rotate_rad(y, x, radians):
    '''
    Rotate by angle in radians

    The rotation follows https://en.wikipedia.org/wiki/Rotation_matrix

    In pixel coordinates with y pointing down and x pointing right,
    the rotation is clockwise. In physical coordinates with y pointing up instead,
    it is counter-clockwise.

    Parameters
    ==========
    y, x : float or numpy.ndarray
        Y and X components of the vector(s)

    radians : float or numpy.ndarray
        Rotation angle in radians

    Returns
    =======
    y, x : float or numpy.ndarray
        Y and X components of rotated vector(s)

    .. versionadded:: 0.6.0
    '''
    return rotate_precalc(y, x, cos_angle=np.cos(radians), sin_angle=np.sin(radians))


def regularize_indices(indices):
    s = indices.shape
    # Output of mgrid
    if (len(s) == 3) and (s[0] == 2):
        result = np.concatenate(indices.T)
    # List of (i, j) pairs
    elif (len(s) == 2) and (s[1] == 2):
        result = indices
    else:
        raise ValueError(
            "Shape of indices is %s, expected (n, 2) or (2, n, m)" % str(indices.shape))
    return result


def frame_peaks(fy, fx, zero, a, b, r, indices):
    indices = regularize_indices(indices)
    peaks = calc_coords(zero, a, b, indices)
    selector = within_frame(peaks, r, fy, fx)
    return indices[selector], peaks[selector]


def calc_coords(zero, a, b, indices):
    '''
    Calculate coordinates from lattice vectors a, b and indices
    '''
    coefficients = np.array((a, b))
    return zero + np.dot(indices, coefficients)


def within_frame(peaks, r, fy, fx):
    '''
    Return a boolean vector indicating peaks that are within (r, r) and (fy - r, fx - r)
    '''
    selector = (peaks >= (r, r)) * (peaks < (fy - r, fx - r))
    return selector.all(axis=-1)



import numpy as np

from libertem.common.math import prod
from libertem.common import Shape, Slice
from libertem.io.dataset.base import _roi_to_nd_indices


def get_coordinates(slice_: Slice, ds_shape: Shape, roi=None) -> np.ndarray:
    """
    Returns `numpy.ndarray` of coordinates that correspond to the frames in the actual
    navigation space which are part of the current tile or partition.

    Parameters
    ----------
    slice_: Slice
        Describes the location within the dataset with navigation
        dimension flattened and reduced to the ROI.

    ds_shape: Shape
        The original shape of the whole dataset, not influenced by the ROI

    roi: numpy.ndarray, optional
        Array of type bool, matching the navigation shape of the dataset
    """
    o = slice_.origin
    s = slice_.shape
    sig_dims = s.sig.dims
    start_idx = o[0]
    end_idx = o[0] + s[0]
    nav_shape = ds_shape[:-sig_dims]
    if roi is None:
        flat_nav_shape = tuple((int(prod(nav_shape)),))
        coordinates = np.stack(
            np.unravel_index(
                np.ravel_multi_index([np.arange(start_idx, end_idx)], flat_nav_shape),
                nav_shape
            ),
            axis=1
        )
    else:
        ds_shape = Shape(ds_shape, sig_dims=sig_dims)
        ds_slice = Slice(origin=[0] * len(ds_shape), shape=ds_shape)
        roi = roi.reshape(nav_shape)
        indices = _roi_to_nd_indices(roi, ds_slice)
        coordinates = np.array(list(indices)[start_idx:end_idx])
    return coordinates



import os
import sys
import json
import random

import numpy as np
import pytest
import cloudpickle

from libertem.io.dataset.k2is import K2ISDataSet, get_filenames
from libertem.executor.inline import InlineJobExecutor
from libertem.analysis.raw import PickFrameAnalysis
from libertem.common.buffers import BufferWrapper
from libertem.udf import UDF
from libertem.udf.raw import PickUDF
from libertem.udf.sumsigudf import SumSigUDF
from libertem.udf.masks import ApplyMasksUDF
from libertem.io.dataset.base import TilingScheme, BufferedBackend, MMapBackend, DirectBackend
from libertem.common import Shape
from libertem.common.buffers import reshaped_view
from libertem import masks

from utils import dataset_correction_verification, get_testdata_path, ValidationUDF, roi_as_sparse


K2IS_TESTDATA_DIR = os.path.join(get_testdata_path(), 'Capture52')
K2IS_TESTDATA_PATH = os.path.join(K2IS_TESTDATA_DIR, 'Capture52_.gtg')
K2IS_TESTDATA_RAW = os.path.join(
    K2IS_TESTDATA_DIR,
    'Capture52_.gtg_(34, 35, 1860, 2048)_uint16.raw'
)
HAVE_K2IS_TESTDATA = os.path.exists(K2IS_TESTDATA_PATH)
HAVE_K2IS_RAWDATA = os.path.exists(K2IS_TESTDATA_RAW)

needsdata = pytest.mark.skipif(not HAVE_K2IS_TESTDATA, reason="need K2IS testdata")  # NOQA


@pytest.fixture
def default_k2is(lt_ctx):
    ds = lt_ctx.load(
        "k2is",
        path=K2IS_TESTDATA_PATH,
        io_backend=MMapBackend(),
    )
    return ds


@pytest.fixture
def buffered_k2is(lt_ctx):
    buffered = BufferedBackend()
    return lt_ctx.load(
        "k2is",
        path=str(K2IS_TESTDATA_PATH),
        io_backend=buffered,
    )


@pytest.fixture
def direct_k2is(lt_ctx):
    direct = DirectBackend()
    return lt_ctx.load(
        "k2is",
        path=str(K2IS_TESTDATA_PATH),
        io_backend=direct,
    )


@pytest.fixture(scope='module')
def default_k2is_raw():
    return np.memmap(
        K2IS_TESTDATA_RAW,
        shape=(34, 35, 1860, 2048),
        dtype=np.uint16,
        mode='r'
    )


def test_glob(tmp_path, naughty_filename):
    naughty_dir = tmp_path / naughty_filename
    naughty_dir.mkdir()

    naughty_gtg = naughty_dir / (naughty_filename + '_.gtg')
    naughty_gtg.touch()

    naughty_bin_0 = naughty_dir / (naughty_filename + '_1.bin')
    naughty_bin_1 = naughty_dir / (naughty_filename + '_2.bin')
    naughty_bin_0.touch()
    naughty_bin_1.touch()

    filenames_1 = set(get_filenames(naughty_gtg))
    filenames_2 = set(get_filenames(naughty_bin_0))
    print(filenames_1)
    print(filenames_2)
    target = {str(naughty_bin_0), str(naughty_bin_1)}
    assert filenames_1 == target
    assert filenames_2 == target


@needsdata
def test_detect():
    params = K2ISDataSet.detect_params(K2IS_TESTDATA_PATH, InlineJobExecutor())["parameters"]
    assert params == {
        "path": K2IS_TESTDATA_PATH,
        "nav_shape": (34, 35),
        "sig_shape": (1860, 2048),
        "sync_offset": 250
    }


@needsdata
def test_simple_open(default_k2is):
    assert tuple(default_k2is.shape) == (34, 35, 1860, 2048)

    # shapes are JSON-encodable:
    json.dumps(tuple(default_k2is.shape))


@needsdata
def test_check_valid(default_k2is):
    assert default_k2is.check_valid()


@needsdata
def test_sync(default_k2is):
    with default_k2is._get_syncer().sectors[0] as sector:
        first_block = next(sector.get_blocks())
    assert first_block.header['frame_id'] == 59


@needsdata
@pytest.mark.slow
def test_read(default_k2is):
    partitions = default_k2is.get_partitions()
    p = next(partitions)
    assert tuple(p.shape)[1:] == (2 * 930, 8 * 256)

    tileshape = Shape(
        (16, 930, 16),
        sig_dims=2,
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=default_k2is.shape,
    )

    tiles = p.get_tiles(tiling_scheme=tiling_scheme)
    t = next(tiles)
    # we get 3D tiles here, because K2IS partitions are inherently 3D
    assert tuple(t.tile_slice.shape) == (16, 930, 16)
    for _ in tiles:
        pass


@needsdata
def test_scheme_too_large(default_k2is):
    partitions = default_k2is.get_partitions()
    p = next(partitions)
    depth = p.shape[0]

    # we make a tileshape that is too large for the partition here:
    tileshape = Shape(
        (depth + 1,) + tuple(default_k2is.shape.sig),
        sig_dims=default_k2is.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=default_k2is.shape,
    )

    # tile shape is clamped to partition shape:
    tiles = p.get_tiles(tiling_scheme=tiling_scheme)
    t = next(tiles)
    assert tuple(t.tile_slice.shape) == tuple((depth,) + default_k2is.shape.sig)
    for _ in tiles:
        pass


@needsdata
@pytest.mark.slow
@pytest.mark.skipif(not HAVE_K2IS_RAWDATA, reason="No K2 IS raw data reference found")
def test_comparison(default_k2is, default_k2is_raw, lt_ctx_fast):
    udf = ValidationUDF(
        reference=reshaped_view(default_k2is_raw, (-1, *tuple(default_k2is.shape.sig))),
        validation_function=lambda a, b: np.all(a == b),
    )
    lt_ctx_fast.run_udf(udf=udf, dataset=default_k2is)


@needsdata
@pytest.mark.slow
@pytest.mark.skipif(not HAVE_K2IS_RAWDATA, reason="No K2 IS raw data reference found")
def test_comparison_roi(default_k2is, default_k2is_raw, lt_ctx_fast):
    roi = np.random.choice(
        [True, False],
        size=tuple(default_k2is.shape.nav),
        p=[0.5, 0.5]
    )
    udf = ValidationUDF(
        reference=default_k2is_raw[roi],
        validation_function=lambda a, b: np.all(a == b),
    )
    lt_ctx_fast.run_udf(udf=udf, dataset=default_k2is, roi=roi)


@needsdata
@pytest.mark.skipif(not HAVE_K2IS_RAWDATA, reason="No K2 IS raw data reference found")
def test_comparison_mask(default_k2is, default_k2is_raw, local_cluster_ctx, lt_ctx):
    default_k2is_raw_ds = local_cluster_ctx.load(
        "raw", K2IS_TESTDATA_RAW, dtype="u2", nav_shape=(34, 35), sig_shape=(1860, 2048),
    )

    udf = ApplyMasksUDF(
        mask_factories=[lambda: masks.circular(centerX=1024, centerY=930, radius=465,
                                               imageSizeX=2048, imageSizeY=1860)]
    )
    r1 = local_cluster_ctx.run_udf(udf=udf, dataset=default_k2is)
    r2 = local_cluster_ctx.run_udf(udf=udf, dataset=default_k2is_raw_ds)
    assert np.allclose(
        r1['intensity'],
        r2['intensity'],
    )


@needsdata
def test_read_full_frames(default_k2is):
    partitions = default_k2is.get_partitions()
    p = next(partitions)
    assert tuple(p.shape)[1:] == (2 * 930, 8 * 256)

    tileshape = Shape(
        (1, 1860, 2048),
        sig_dims=2,
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=default_k2is.shape,
    )

    tiles = p.get_tiles(tiling_scheme=tiling_scheme)
    t = next(tiles)
    assert tuple(t.tile_slice.shape) == (1, 1860, 2048)
    assert tuple(t.tile_slice.origin) == (0, 0, 0)

    for t in tiles:
        assert t.tile_slice.origin[0] < p.shape[0]


@needsdata
def test_read_invalid_tileshape(default_k2is):
    partitions = default_k2is.get_partitions()
    p = next(partitions)

    tileshape = Shape(
        (1, 930, 10),
        sig_dims=2,
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=default_k2is.shape,
    )

    with pytest.raises(ValueError):
        next(p.get_tiles(tiling_scheme=tiling_scheme))


@needsdata
@pytest.mark.slow
def test_apply_mask_analysis(default_k2is, local_cluster_ctx):
    mask = np.ones((1860, 2048))
    analysis = local_cluster_ctx.create_mask_analysis(
        factories=[lambda: mask], dataset=default_k2is,
    )
    results = local_cluster_ctx.run(analysis)
    assert results[0].raw_data.shape == (34, 35)


@needsdata
def test_sum_analysis(default_k2is, local_cluster_ctx):
    analysis = local_cluster_ctx.create_sum_analysis(dataset=default_k2is)
    results = local_cluster_ctx.run(analysis)
    assert results[0].raw_data.shape == (1860, 2048)


@needsdata
def test_pick_analysis(default_k2is, local_cluster_ctx):
    analysis = PickFrameAnalysis(dataset=default_k2is, parameters={"x": 16, "y": 16})
    results = local_cluster_ctx.run(analysis)
    assert results[0].raw_data.shape == (1860, 2048)


@needsdata
@pytest.mark.parametrize(
    # Default is too large for test without ROI
    "with_roi", (True, )
)
@pytest.mark.slow
def test_correction(default_k2is, lt_ctx, with_roi):
    ds = default_k2is

    if with_roi:
        roi = np.zeros(ds.shape.nav, dtype=bool)
        roi[:1] = True
    else:
        roi = None

    dataset_correction_verification(ds=ds, roi=roi, lt_ctx=lt_ctx)


@needsdata
def test_dataset_is_picklable(default_k2is):
    pickled = cloudpickle.dumps(default_k2is)
    cloudpickle.loads(pickled)

    # let's keep the pickled dataset size small-ish:
    assert len(pickled) < 4 * 1024


@needsdata
def test_partition_is_picklable(default_k2is):
    pickled = cloudpickle.dumps(next(default_k2is.get_partitions()))
    cloudpickle.loads(pickled)

    # let's keep the pickled dataset size small-ish:
    assert len(pickled) < 4 * 1024


@needsdata
def test_get_diags(default_k2is):
    diags = default_k2is.diagnostics

    # diags are JSON-encodable:
    json.dumps(diags)


@needsdata
@pytest.mark.slow
def test_udf_on_k2is(lt_ctx, default_k2is):
    res = lt_ctx.map(
        dataset=default_k2is,
        f=np.sum,
    )
    res.data
    res.raw_data
    # print(data.shape, res['pixelsum'].data.shape)
    # assert np.allclose(res['pixelsum'].data, np.sum(data, axis=(2, 3)))


class PixelsumUDF(UDF):
    def get_result_buffers(self):
        return {
            'pixelsum': BufferWrapper(
                kind="nav", dtype="float32"
            )
        }

    def process_frame(self, frame):
        assert self.results.pixelsum.shape == (1,)
        self.results.pixelsum[:] = np.sum(frame)


@needsdata
@pytest.mark.with_numba
def test_udf_roi(lt_ctx, default_k2is):
    roi = np.zeros(default_k2is.shape.flatten_nav().nav, dtype=bool)
    roi[0] = 1
    psum = PixelsumUDF()
    res = lt_ctx.run_udf(dataset=default_k2is, udf=psum, roi=roi)
    assert 'pixelsum' in res


@needsdata
def test_roi(lt_ctx, default_k2is):
    p = next(default_k2is.get_partitions())
    roi = np.zeros(p.shape.flatten_nav().nav, dtype=bool)
    roi[0] = 1
    tiles = []

    tileshape = Shape(
        (16, 930, 16),
        sig_dims=2,
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=default_k2is.shape,
    )

    for tile in p.get_tiles(dest_dtype="float32", roi=roi, tiling_scheme=tiling_scheme):
        print("tile:", tile)
        tiles.append(tile)
    assert len(tiles) == 2*8*16


@needsdata
def test_macrotile_normal(lt_ctx, default_k2is):
    ps = default_k2is.get_partitions()
    _ = next(ps)
    p2 = next(ps)
    macrotile = p2.get_macrotile()
    assert macrotile.tile_slice.shape == p2.shape
    assert macrotile.tile_slice.origin[0] == p2._start_frame


@needsdata
def test_macrotile_roi_1(lt_ctx, default_k2is):
    roi = np.zeros(default_k2is.shape.nav, dtype=bool)
    roi[0, 5] = 1
    roi[0, 17] = 1
    p = next(default_k2is.get_partitions())
    macrotile = p.get_macrotile(roi=roi)
    assert tuple(macrotile.tile_slice.shape) == (2, 1860, 2048)


@needsdata
def test_macrotile_roi_2(lt_ctx, default_k2is):
    roi = np.zeros(default_k2is.shape.nav, dtype=bool)
    # all ones are in the first partition, so we don't get any data in p2:
    roi[0, 5] = 1
    roi[0, 17] = 1
    ps = default_k2is.get_partitions()
    _ = next(ps)
    p2 = next(ps)
    macrotile = p2.get_macrotile(roi=roi)
    assert tuple(macrotile.tile_slice.shape) == (0, 1860, 2048)


@needsdata
def test_macrotile_roi_3(lt_ctx, default_k2is):
    roi = np.ones(default_k2is.shape.nav, dtype=bool)
    ps = default_k2is.get_partitions()
    _ = next(ps)
    p2 = next(ps)
    macrotile = p2.get_macrotile(roi=roi)
    assert tuple(macrotile.tile_slice.shape) == tuple(p2.shape)


@needsdata
def test_cache_key_json_serializable(default_k2is):
    json.dumps(default_k2is.get_cache_key())


@needsdata
@pytest.mark.dist
def test_k2is_dist(dist_ctx):
    ds = K2ISDataSet(path=K2IS_TESTDATA_PATH)
    import glob
    print(dist_ctx.executor.run_function(lambda: os.listdir(K2IS_TESTDATA_DIR)))
    print(dist_ctx.executor.run_function(lambda: list(sorted(glob.glob(K2IS_TESTDATA_DIR+'/*')))))
    ds = ds.initialize(dist_ctx.executor)
    roi = np.zeros(ds.shape.nav, dtype=bool)
    roi[0, 5] = 1
    roi[0, 17] = 1
    analysis = dist_ctx.create_sum_analysis(dataset=ds)
    results = dist_ctx.run(analysis, roi=roi)
    assert results[0].raw_data.shape == (1860, 2048)


@needsdata
def test_compare_backends(lt_ctx, default_k2is, buffered_k2is):
    y = random.choice(range(default_k2is.shape.nav[0]))
    x = random.choice(range(default_k2is.shape.nav[1]))
    mm_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=default_k2is,
        x=x, y=y,
    )).intensity
    buffered_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=buffered_k2is,
        x=x, y=y,
    )).intensity

    assert np.allclose(mm_f0, buffered_f0)


@needsdata
@pytest.mark.skipif(
    sys.platform.startswith("darwin"),
    reason="No support for direct I/O on Mac OS X"
)
def test_compare_direct_to_mmap(lt_ctx, default_k2is, direct_k2is):
    y = random.choice(range(default_k2is.shape.nav[0]))
    x = random.choice(range(default_k2is.shape.nav[1]))
    mm_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=default_k2is,
        x=x, y=y,
    )).intensity
    buffered_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=direct_k2is,
        x=x, y=y,
    )).intensity

    assert np.allclose(mm_f0, buffered_f0)


@needsdata
@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
def test_compare_backends_sparse(lt_ctx, default_k2is, buffered_k2is, as_sparse):
    roi = np.zeros(default_k2is.shape.nav, dtype=bool).reshape((-1,))
    roi[0] = True
    roi[1] = True
    roi[16] = True
    roi[32] = True
    roi[-1] = True
    if as_sparse:
        roi = roi_as_sparse(roi)
    mm_f0 = lt_ctx.run_udf(dataset=default_k2is, udf=PickUDF(), roi=roi)['intensity']
    buffered_f0 = lt_ctx.run_udf(dataset=buffered_k2is, udf=PickUDF(), roi=roi)['intensity']

    assert np.allclose(mm_f0, buffered_f0)


@needsdata
@pytest.mark.with_numba
def test_regression_simple_stride(lt_ctx, default_k2is):
    # bug that only seems happens if tileshape[-1] == 16 and tileshape[1] != 930:
    # <TilingScheme (depth=6) shapes=[(1860, 16)] len=128>
    ts = TilingScheme.make_for_shape(
        tileshape=Shape((6, 1860, 16), sig_dims=2),
        dataset_shape=default_k2is.shape.flatten_nav()
    )
    print(ts)
    p = list(default_k2is.get_partitions())[-1]
    tiles = p.get_tiles(tiling_scheme=ts)
    next(tiles)
    for _ in tiles:
        pass


@needsdata
def test_positive_sync_offset_1(default_k2is, lt_ctx):
    udf = PickUDF()
    # native_sync_offset is 250
    sync_offset = 252

    roi = np.zeros(default_k2is.shape.nav, dtype=bool)
    flat_roi = reshaped_view(roi, -1)
    flat_roi[2:10] = True

    ref = lt_ctx.run_udf(dataset=default_k2is, udf=udf, roi=roi)

    ds = lt_ctx.load(
        "k2is", path=K2IS_TESTDATA_PATH, nav_shape=(4, 2), sync_offset=sync_offset,
    )
    result = lt_ctx.run_udf(dataset=ds, udf=udf)

    assert np.allclose(
        result['intensity'].raw_data,
        ref['intensity'].raw_data
    )


@needsdata
def test_positive_sync_offset_2(default_k2is, lt_ctx):
    udf = PickUDF()
    # native_sync_offset is 250
    sync_offset = 2

    roi_1 = np.zeros(default_k2is.shape.nav, dtype=bool)
    flat_roi_1 = reshaped_view(roi_1, -1)
    flat_roi_1[:8] = True

    ref = lt_ctx.run_udf(dataset=default_k2is, udf=udf, roi=roi_1)

    roi_2 = np.zeros(default_k2is.shape.nav, dtype=bool)
    flat_roi_2 = reshaped_view(roi_2, -1)
    flat_roi_2[248:256] = True

    ds = lt_ctx.load(
        "k2is", path=K2IS_TESTDATA_PATH, sync_offset=sync_offset
    )
    result = lt_ctx.run_udf(dataset=ds, udf=udf, roi=roi_2)

    assert np.allclose(
        result['intensity'].raw_data,
        ref['intensity'].raw_data
    )


@needsdata
def test_zero_sync_offset(default_k2is, lt_ctx):
    udf = PickUDF()
    # native_sync_offset is 250
    sync_offset = 0

    roi_1 = np.zeros(default_k2is.shape.nav, dtype=bool)
    flat_roi_1 = reshaped_view(roi_1, -1)
    flat_roi_1[:8] = True

    ref = lt_ctx.run_udf(dataset=default_k2is, udf=udf, roi=roi_1)

    roi_2 = np.zeros(default_k2is.shape.nav, dtype=bool)
    flat_roi_2 = reshaped_view(roi_2, -1)
    flat_roi_2[250:258] = True

    ds = lt_ctx.load(
        "k2is", path=K2IS_TESTDATA_PATH, sync_offset=sync_offset
    )
    result = lt_ctx.run_udf(dataset=ds, udf=udf, roi=roi_2)

    assert np.allclose(
        result['intensity'].raw_data,
        ref['intensity'].raw_data
    )


@needsdata
def test_negative_sync_offset(default_k2is, lt_ctx):
    udf = PickUDF()
    # native_sync_offset is 250
    sync_offset = -2

    roi_1 = np.zeros(default_k2is.shape.nav, dtype=bool)
    flat_roi_1 = reshaped_view(roi_1, -1)
    flat_roi_1[:8] = True

    ref = lt_ctx.run_udf(dataset=default_k2is, udf=udf, roi=roi_1)

    roi_2 = np.zeros(default_k2is.shape.nav, dtype=bool)
    flat_roi_2 = reshaped_view(roi_2, -1)
    flat_roi_2[252:260] = True

    ds = lt_ctx.load(
        "k2is", path=K2IS_TESTDATA_PATH, sync_offset=sync_offset
    )
    result = lt_ctx.run_udf(dataset=ds, udf=udf, roi=roi_2)

    assert np.allclose(
        result['intensity'].raw_data,
        ref['intensity'].raw_data
    )


@needsdata
def test_offset_smaller_than_image_count(lt_ctx):
    sync_offset = -1520

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "k2is",
            path=K2IS_TESTDATA_PATH,
            sync_offset=sync_offset
        )
    assert e.match(
        r"offset should be in \(-1517, 1517\), which is \(-image_count, image_count\)"
    )


@needsdata
def test_offset_greater_than_image_count(lt_ctx):
    sync_offset = 1520

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "k2is",
            path=K2IS_TESTDATA_PATH,
            sync_offset=sync_offset
        )
    assert e.match(
        r"offset should be in \(-1517, 1517\), which is \(-image_count, image_count\)"
    )


@needsdata
def test_reshape_nav(default_k2is, lt_ctx):
    udf = PickUDF()

    roi = np.zeros(default_k2is.shape.nav, dtype=bool)
    flat_roi = reshaped_view(roi, -1)
    flat_roi[:8] = True

    ref = lt_ctx.run_udf(dataset=default_k2is, udf=udf, roi=roi)

    ds_1 = lt_ctx.load(
        "k2is",
        path=K2IS_TESTDATA_PATH,
        nav_shape=(8,),
    )
    result_1 = lt_ctx.run_udf(dataset=ds_1, udf=udf)
    shape_1 = lt_ctx.run_udf(dataset=ds_1, udf=SumSigUDF())

    assert shape_1['intensity'].data.shape == (8, )
    assert np.allclose(result_1['intensity'].raw_data, ref['intensity'].raw_data)

    ds_2 = lt_ctx.load(
        "k2is",
        path=K2IS_TESTDATA_PATH,
        nav_shape=(2, 2, 2),
    )
    result_2 = lt_ctx.run_udf(dataset=ds_2, udf=udf)
    shape_2 = lt_ctx.run_udf(dataset=ds_2, udf=SumSigUDF())
    assert shape_2['intensity'].data.shape == (2, 2, 2)
    assert np.allclose(result_2['intensity'].raw_data, ref['intensity'].raw_data)


@needsdata
def test_incorrect_sig_shape(lt_ctx):
    sig_shape = (5, 5)

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "k2is",
            path=K2IS_TESTDATA_PATH,
            sig_shape=sig_shape
        )
    assert e.match(
        r"sig_shape must be of size: 3809280"
    )


def test_bad_params(ds_params_tester, standard_bad_ds_params):
    args = ("k2is", K2IS_TESTDATA_PATH)
    for params in standard_bad_ds_params:
        ds_params_tester(*args, **params)



import os
import sys
import json
import random

import pytest
import numpy as np
import cloudpickle

from libertem.executor.inline import InlineJobExecutor
from libertem.analysis.raw import PickFrameAnalysis
from libertem.io.dataset.base import (
    DataSetException, TilingScheme, BufferedBackend, MMapBackend, DirectBackend
)
from libertem.io.dataset.empad import EMPADDataSet, get_params_from_xml
from libertem.common import Shape
from libertem.common.buffers import reshaped_view
from libertem.udf.sumsigudf import SumSigUDF
from libertem.udf.raw import PickUDF
from utils import _mk_random

from utils import dataset_correction_verification, get_testdata_path, ValidationUDF, roi_as_sparse

EMPAD_TESTDATA_PATH = os.path.join(get_testdata_path(), 'EMPAD')
EMPAD_RAW = os.path.join(EMPAD_TESTDATA_PATH, 'scan_11_x4_y4.raw')
EMPAD_XML = os.path.join(EMPAD_TESTDATA_PATH, 'acquisition_12_pretty.xml')
EMPAD_XML_2 = os.path.join(EMPAD_TESTDATA_PATH, 'acquisition_12_pretty.xml')
EMPAD_XML_SERIES = os.path.join(EMPAD_TESTDATA_PATH, 'acquisition_series_pretty.xml')
EMPAD_XML_BROKEN_PARAMS = os.path.join(
    EMPAD_TESTDATA_PATH, 'Magical_DataSet/acquisition_1/acquisition_1.xml',
)
EMPAD_XML_BROKEN_PARAMS_BAD = os.path.join(
    EMPAD_TESTDATA_PATH, 'Magical_DataSet/acquisition_1/acquisition_1_bad.xml',
)
HAVE_EMPAD_TESTDATA = os.path.exists(EMPAD_RAW) and os.path.exists(EMPAD_XML)

pytestmark = pytest.mark.skipif(not HAVE_EMPAD_TESTDATA, reason="need EMPAD testdata")  # NOQA


@pytest.fixture
def default_empad(lt_ctx):
    ds = lt_ctx.load(
        "empad",
        path=EMPAD_XML,
        io_backend=MMapBackend(),
    )
    yield ds


@pytest.fixture
def buffered_empad(lt_ctx):
    buffered = BufferedBackend()
    return lt_ctx.load(
        "empad",
        path=EMPAD_XML,
        io_backend=buffered,
    )


@pytest.fixture
def direct_empad(lt_ctx):
    direct = DirectBackend()
    return lt_ctx.load(
        "empad",
        path=EMPAD_XML,
        io_backend=direct,
    )


@pytest.fixture(scope='module')
def default_empad_raw():
    raw_data = np.memmap(
        EMPAD_RAW,
        shape=(4, 4, 130, 128),
        dtype=np.float32,
        mode='r'
    )
    return raw_data[:, :, :128, :]


def test_new_empad_xml(lt_ctx):
    ds = lt_ctx.load("empad", path=EMPAD_XML_2)
    assert ds.shape.to_tuple() == (4, 4, 128, 128)


def test_empad_broken_scan_parameters(lt_ctx):
    ds = lt_ctx.load("empad", path=EMPAD_XML_BROKEN_PARAMS)
    assert ds.shape.to_tuple() == (128, 16, 128, 128)


def test_empad_broken_scan_parameters_bad(lt_ctx):
    with pytest.raises(ValueError) as excinfo:
        lt_ctx.load("empad", path=EMPAD_XML_BROKEN_PARAMS_BAD)
    assert "(128, 17)" in str(excinfo.value)
    assert "(256, 256)" in str(excinfo.value)


def test_series_acquisition_xml():
    path_raw, nav_shape = get_params_from_xml(EMPAD_XML_SERIES)
    assert nav_shape == (1000,)


@pytest.fixture(scope='session')
def random_empad(tmpdir_factory):
    executor = InlineJobExecutor()
    datadir = tmpdir_factory.mktemp('data')
    filename = datadir + '/empad-test-default.raw'
    data = _mk_random(size=(16, 16, 130, 128), dtype='float32')
    data.tofile(str(filename))
    del data
    ds = EMPADDataSet(
        path=str(filename),
        nav_shape=(16, 16),
    )
    ds = ds.initialize(executor)
    yield ds


def test_simple_open(default_empad):
    assert tuple(default_empad.shape) == (4, 4, 128, 128)


def test_check_valid(default_empad):
    assert default_empad.check_valid()


def test_check_valid_random(random_empad):
    assert random_empad.check_valid()


def test_comparison(default_empad, default_empad_raw, lt_ctx_fast):
    udf = ValidationUDF(
        reference=reshaped_view(default_empad_raw, (-1, *tuple(default_empad.shape.sig)))
    )
    lt_ctx_fast.run_udf(udf=udf, dataset=default_empad)


def test_comparison_roi(default_empad, default_empad_raw, lt_ctx_fast):
    roi = np.random.choice(
        [True, False],
        size=tuple(default_empad.shape.nav),
        p=[0.5, 0.5]
    )
    udf = ValidationUDF(reference=default_empad_raw[roi])
    lt_ctx_fast.run_udf(udf=udf, dataset=default_empad, roi=roi)


def test_read_random(random_empad):
    partitions = random_empad.get_partitions()
    p = next(partitions)
    # FIXME: partition shape can vary by number of cores
    # assert tuple(p.shape) == (2, 16, 128, 128)
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=Shape((16, 128, 128), sig_dims=2),
        dataset_shape=random_empad.shape,
    )
    tiles = p.get_tiles(tiling_scheme=tiling_scheme)
    t = next(tiles)

    # ~1MB
    assert tuple(t.tile_slice.shape) == (16, 128, 128)

    for _ in tiles:
        pass


def test_read(default_empad):
    partitions = default_empad.get_partitions()
    p = next(partitions)
    # FIXME: partition shape can vary by number of cores
    # assert tuple(p.shape) == (2, 16, 128, 128)
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=Shape((16, 128, 128), sig_dims=2),
        dataset_shape=default_empad.shape,
    )
    tiles = p.get_tiles(tiling_scheme=tiling_scheme)
    t = next(tiles)

    # ~1MB
    assert tuple(t.tile_slice.shape) == (16, 128, 128)

    for _ in tiles:
        pass


def test_scheme_too_large(default_empad):
    partitions = default_empad.get_partitions()
    p = next(partitions)
    depth = p.shape[0]

    # we make a tileshape that is too large for the partition here:
    tileshape = Shape(
        (depth + 1,) + tuple(default_empad.shape.sig),
        sig_dims=default_empad.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=default_empad.shape,
    )

    # tile shape is clamped to partition shape:
    tiles = p.get_tiles(tiling_scheme=tiling_scheme)
    t = next(tiles)
    assert tuple(t.tile_slice.shape) == tuple((depth,) + default_empad.shape.sig)

    for _ in tiles:
        pass


def test_pickle_is_small(default_empad):
    pickled = cloudpickle.dumps(default_empad)
    cloudpickle.loads(pickled)

    # let's keep the pickled dataset size small-ish:
    assert len(pickled) < 2 * 1024


def test_apply_mask_analysis(default_empad, lt_ctx):
    mask = np.ones((128, 128))
    analysis = lt_ctx.create_mask_analysis(factories=[lambda: mask], dataset=default_empad)
    results = lt_ctx.run(analysis)
    assert results[0].raw_data.shape == (4, 4)
    assert np.count_nonzero(results[0].raw_data) > 0


def test_sum_analysis(default_empad, lt_ctx):
    analysis = lt_ctx.create_sum_analysis(dataset=default_empad)
    results = lt_ctx.run(analysis)
    assert results[0].raw_data.shape == (128, 128)
    assert np.count_nonzero(results[0].raw_data) > 0


def test_pick_analysis(default_empad, lt_ctx):
    analysis = PickFrameAnalysis(dataset=default_empad, parameters={"x": 2, "y": 2})
    results = lt_ctx.run(analysis)
    assert results[0].raw_data.shape == (128, 128)
    assert np.count_nonzero(results[0].raw_data) > 0


@pytest.mark.parametrize(
    "with_roi", (True, False)
)
def test_correction(default_empad, lt_ctx, with_roi):
    ds = default_empad

    if with_roi:
        roi = np.zeros(ds.shape.nav, dtype=bool)
        roi[:1] = True
    else:
        roi = None

    dataset_correction_verification(ds=ds, roi=roi, lt_ctx=lt_ctx)


def test_nonexistent():
    ds = EMPADDataSet(
        path="/does/not/exist.raw",
        nav_shape=(4, 4),
    )
    with pytest.raises(DataSetException) as einfo:
        ds = ds.initialize(InlineJobExecutor())
    assert einfo.match("could not open file /does/not/exist.raw")


def test_detect_fail():
    executor = InlineJobExecutor()
    # does not exist:
    assert not EMPADDataSet.detect_params("/does/not/exist.raw", executor=executor)
    # exists but we can't detect any parameters (and we don't know if it even is an EMPAD file)
    assert not EMPADDataSet.detect_params(EMPAD_RAW, executor=executor)


def test_cache_key_json_serializable(default_empad):
    json.dumps(default_empad.get_cache_key())


@pytest.mark.dist
def test_empad_dist(dist_ctx):
    ds = EMPADDataSet(path=EMPAD_XML)
    ds = ds.initialize(dist_ctx.executor)
    analysis = dist_ctx.create_sum_analysis(dataset=ds)
    results = dist_ctx.run(analysis)
    assert results[0].raw_data.shape == (128, 128)


def test_positive_sync_offset(default_empad, lt_ctx):
    udf = SumSigUDF()
    sync_offset = 2

    ds_with_offset = lt_ctx.load(
        "empad", path=EMPAD_XML, sync_offset=sync_offset
    )

    result = lt_ctx.run_udf(dataset=default_empad, udf=udf)
    result = result['intensity'].raw_data[sync_offset:]

    result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf)
    result_with_offset = result_with_offset['intensity'].raw_data[
        :ds_with_offset._meta.image_count - sync_offset
    ]

    assert np.allclose(result, result_with_offset)


def test_negative_sync_offset(default_empad, lt_ctx):
    udf = SumSigUDF()
    sync_offset = -2

    ds_with_offset = lt_ctx.load(
        "empad", path=EMPAD_XML, sync_offset=sync_offset
    )

    result = lt_ctx.run_udf(dataset=default_empad, udf=udf)
    result = result['intensity'].raw_data[:default_empad._meta.image_count - abs(sync_offset)]

    result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf)
    result_with_offset = result_with_offset['intensity'].raw_data[abs(sync_offset):]

    assert np.allclose(result, result_with_offset)


def test_positive_sync_offset_with_roi(default_empad, lt_ctx):
    udf = SumSigUDF()
    result = lt_ctx.run_udf(dataset=default_empad, udf=udf)
    result = result['intensity'].raw_data

    sync_offset = 2

    ds_with_offset = lt_ctx.load(
        "empad", path=EMPAD_XML, sync_offset=sync_offset
    )

    roi = np.random.choice([False], (4, 4))
    roi[0:1] = True

    result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf, roi=roi)
    result_with_offset = result_with_offset['intensity'].raw_data

    assert np.allclose(result[sync_offset:4 + sync_offset], result_with_offset)


def test_negative_sync_offset_with_roi(default_empad, lt_ctx):
    udf = SumSigUDF()
    result = lt_ctx.run_udf(dataset=default_empad, udf=udf)
    result = result['intensity'].raw_data

    sync_offset = -2

    ds_with_offset = lt_ctx.load(
        "empad", path=EMPAD_XML, sync_offset=sync_offset
    )

    roi = np.random.choice([False], (4, 4))
    roi[0:1] = True

    result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf, roi=roi)
    result_with_offset = result_with_offset['intensity'].raw_data

    assert np.allclose(result[:4 + sync_offset], result_with_offset[abs(sync_offset):])


def test_offset_smaller_than_image_count(lt_ctx):
    sync_offset = -20

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "empad",
            path=EMPAD_XML,
            sync_offset=sync_offset
        )
    assert e.match(
        r"offset should be in \(-16, 16\), which is \(-image_count, image_count\)"
    )


def test_offset_greater_than_image_count(lt_ctx):
    sync_offset = 20

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "empad",
            path=EMPAD_XML,
            sync_offset=sync_offset
        )
    assert e.match(
        r"offset should be in \(-16, 16\), which is \(-image_count, image_count\)"
    )


def test_reshape_nav(lt_ctx):
    udf = SumSigUDF()

    ds_with_1d_nav = lt_ctx.load("empad", path=EMPAD_XML, nav_shape=(8,))
    result_with_1d_nav = lt_ctx.run_udf(dataset=ds_with_1d_nav, udf=udf)
    result_with_1d_nav = result_with_1d_nav['intensity'].raw_data

    ds_with_2d_nav = lt_ctx.load("empad", path=EMPAD_XML, nav_shape=(4, 2))
    result_with_2d_nav = lt_ctx.run_udf(dataset=ds_with_2d_nav, udf=udf)
    result_with_2d_nav = result_with_2d_nav['intensity'].raw_data

    ds_with_3d_nav = lt_ctx.load("empad", path=EMPAD_XML, nav_shape=(2, 2, 2))
    result_with_3d_nav = lt_ctx.run_udf(dataset=ds_with_3d_nav, udf=udf)
    result_with_3d_nav = result_with_3d_nav['intensity'].raw_data

    assert np.allclose(result_with_1d_nav, result_with_2d_nav, result_with_3d_nav)


@pytest.mark.parametrize(
    "sig_shape", ((16384,), (128, 8, 16))
)
def test_reshape_sig(lt_ctx, default_empad, sig_shape):
    udf = SumSigUDF()

    result = lt_ctx.run_udf(dataset=default_empad, udf=udf)
    result = result['intensity'].raw_data

    ds_1 = lt_ctx.load("empad", path=EMPAD_XML, sig_shape=sig_shape)
    result_1 = lt_ctx.run_udf(dataset=ds_1, udf=udf)
    result_1 = result_1['intensity'].raw_data

    assert np.allclose(result, result_1)


def test_incorrect_sig_shape(lt_ctx):
    sig_shape = (5, 5)

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "empad",
            path=EMPAD_XML,
            sig_shape=sig_shape
        )
    assert e.match(
        r"sig_shape must be of size: 16384"
    )


def test_scan_size_deprecation(lt_ctx):
    scan_size = (2, 2)

    with pytest.warns(FutureWarning):
        ds = lt_ctx.load(
            "empad",
            path=EMPAD_XML,
            scan_size=scan_size,
        )
    assert tuple(ds.shape) == (2, 2, 128, 128)


def test_compare_backends(lt_ctx, default_empad, buffered_empad):
    y = random.choice(range(default_empad.shape.nav[0]))
    x = random.choice(range(default_empad.shape.nav[1]))
    mm_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=default_empad,
        x=x, y=y,
    )).intensity
    buffered_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=buffered_empad,
        x=x, y=y,
    )).intensity

    assert np.allclose(mm_f0, buffered_f0)


@pytest.mark.skipif(
    sys.platform.startswith("darwin"),
    reason="No support for direct I/O on Mac OS X"
)
def test_compare_direct_to_mmap(lt_ctx, default_empad, direct_empad):
    y = random.choice(range(default_empad.shape.nav[0]))
    x = random.choice(range(default_empad.shape.nav[1]))
    mm_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=default_empad,
        x=x, y=y,
    )).intensity
    buffered_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=direct_empad,
        x=x, y=y,
    )).intensity

    assert np.allclose(mm_f0, buffered_f0)


@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
def test_compare_backends_sparse(lt_ctx, default_empad, buffered_empad, as_sparse):
    roi = np.zeros(default_empad.shape.nav, dtype=bool).reshape((-1,))
    roi[0] = True
    roi[1] = True
    roi[8] = True
    roi[-1] = True
    if as_sparse:
        roi = roi_as_sparse(roi)
    mm_f0 = lt_ctx.run_udf(dataset=default_empad, udf=PickUDF(), roi=roi)['intensity']
    buffered_f0 = lt_ctx.run_udf(dataset=buffered_empad, udf=PickUDF(), roi=roi)['intensity']

    assert np.allclose(mm_f0, buffered_f0)


def test_bad_params(ds_params_tester, standard_bad_ds_params):
    args = ("empad", EMPAD_XML)
    for params in standard_bad_ds_params:
        ds_params_tester(*args, **params)



from typing import NamedTuple
import numpy as np

from libertem.common.math import prod


class OffsetsSizes(NamedTuple):
    """
    General slicing for slicing the data out of a file
    (removing file header and per-frame headers)
    """

    # whole-file slicing, in bytes
    file_offset: int  # number of bytes to skip at the beginning of the file
    skip_end: int  # how many bytes to skip at the end

    # per-frame slicing, in number of items in native dtype
    frame_offset: int  # number of items to skip at the beginning of each frame
    frame_size: int  # number of items for each frame


class File:
    """
    A description of a file that is part of a dataset. Contains
    information about the internal structure, like sizes
    of headers, frames, frame headers, frame footers, ...

    Parameters
    ----------
    path : str
        The path of the file. Interpretation may be backend-specific

    start_idx : int
        Start index of signal elements in this file (inclusive),
        in the flattened navigation axis

    end_idx : int
        End index of signal elements in this file (exclusive),
        in the flattened navigation axis

    native_dtype : np.dtype
        The dtype that is used for reading the data. This may
        match the "real" dtype of data, or in some cases, when
        no direct match is possible (decoding is necessary),
        it falls back to bytes (np.uint8)

    sig_shape : Shape | Tuple[int, ...]
        The shape of each signal element

    file_header: int
        Number of bytes to ignore at the beginning of the file

    frame_header: int
        Number of bytes to ignore before each frame

    frame_footer: int
        Number of bytes to ignore after each frame
    """
    def __init__(self, path, start_idx, end_idx,
                native_dtype, sig_shape,
                frame_footer=0, frame_header=0, file_header=0):
        self._start_idx = start_idx
        self._end_idx = end_idx
        self._native_dtype = native_dtype
        self._path = path
        self._file_header = file_header
        self._frame_header = frame_header
        self._frame_footer = frame_footer
        self._sig_shape = tuple(sig_shape)

    @property
    def file_header_bytes(self) -> int:
        return self._file_header

    @property
    def start_idx(self) -> int:
        return self._start_idx

    @property
    def end_idx(self) -> int:
        return self._end_idx

    @property
    def num_frames(self) -> int:
        return self._end_idx - self._start_idx

    @property
    def sig_shape(self) -> tuple[int, ...]:
        return self._sig_shape

    @property
    def native_dtype(self) -> np.dtype:
        return self._native_dtype

    @property
    def path(self) -> str:
        return self._path

    def get_offsets_sizes(self, size: int) -> OffsetsSizes:
        """
        Get file and frame offsets/sizes

        Parameters
        ----------
        size : int
            len(memoryview) for the whole file

        Returns
        -------
        slicing
            The file/frame slicing
        """
        itemsize = np.dtype(self._native_dtype).itemsize
        assert self._frame_header % itemsize == 0
        assert self._frame_footer % itemsize == 0
        frame_size = int(prod(self._sig_shape))
        frame_offset = self._frame_header // itemsize
        file_offset = self._file_header
        skip_end = 0

        # cut off any extra data at the end of the file:
        if size % int(prod(self._sig_shape)):
            new_mmap_size = self.num_frames * (
                (itemsize * frame_size) + self._frame_header + self._frame_footer
            )
            skip_end = (size - file_offset) - new_mmap_size
        assert skip_end >= 0

        return OffsetsSizes(
            file_offset=file_offset,
            skip_end=skip_end,
            frame_offset=frame_offset,
            frame_size=frame_size,
        )

    def get_array_from_memview(self, mem: memoryview, slicing: OffsetsSizes) -> np.ndarray:
        """
        Convert a memoryview of the file's data into an ndarray, cutting away
        frame headers and footers as defined by `start` and `stop` parameters.

        Parameters
        ----------
        mem
            The input memoryview

        start
            Cut off frame headers of this size; usually start = frame_header_bytes // itemsize

        stop
            End index; usually stop = start + prod(sig_shape)

        Returns
        -------
        np.ndarray
            The output array. Should have shape (num_frames, prod(sig_shape)) and native dtype
        """
        if slicing.skip_end > 0:
            mem = mem[slicing.file_offset:-slicing.skip_end]
        else:
            mem = mem[slicing.file_offset:]
        itemsize = np.dtype(self._native_dtype).itemsize
        assert len(mem) % itemsize == 0, \
            "len(mem) must fit the dtype"
        assert len(mem) // itemsize % self.num_frames == 0, \
            "len(mem) must fit the number of frames"
        assert len(mem) // itemsize // self.num_frames % (
            slicing.frame_size + (self._frame_header + self._frame_footer) // itemsize
        ) == 0, "len(mem) must fit the sig shape"
        arr_uncut = np.frombuffer(mem, dtype=self._native_dtype).reshape(
            (self.num_frames, -1)
        )
        arr = arr_uncut[:, slicing.frame_offset:slicing.frame_offset + slicing.frame_size]
        assert arr.shape[1] == slicing.frame_size, \
            "array shape must fit the signal shape"
        assert arr.size > 0
        return arr

    def __repr__(self):
        return "<%s %d:%d>" % (self.__class__.__name__, self._start_idx, self._end_idx)



from libertem.common.analysis import AnalysisResultSet


class ResultFormatRegistry(type):
    """
    This is a metaclass to make ResultFormat subclasses register
    themselves. On import of the module the subclass is implemented in,
    the __init__ method is called, which in turn asks the class
    about their ID, a description, ...
    """

    # FIXME: TypedDict, needs typing-extensions
    registry: dict = {}

    def __init__(cls, name, args, bases):
        format_info = cls.format_info()
        if format_info is None:
            return
        cls.registry[format_info["id"]] = {
            "class": cls,
            "info": format_info,
        }

    @classmethod
    def get_available_formats(cls):
        return {
            identifier: {
                "identifier": identifier,
                "description": fmt['info']['description'],
            }
            for identifier, fmt in cls.registry.items()
        }

    @classmethod
    def get_format_by_id(cls, identifier) -> type['ResultFormat']:
        return cls.registry[identifier]['class']


class ResultFormat(metaclass=ResultFormatRegistry):
    @classmethod
    def format_info(cls):
        return None

    def __init__(self, result_set: AnalysisResultSet):
        self._result_set = result_set

    def serialize_to_buffer(self, buf):
        """
        Serialize to `buf`, which is a file-like, i.e. `io.BytesIO`
        """
        raise NotImplementedError()

    def get_result_keys(self):
        for k in self._result_set.keys():
            if not self._result_set[k].include_in_download:
                continue
            yield k

    def get_content_type(self):
        raise NotImplementedError()

    def get_filename(self):
        raise NotImplementedError()



import numpy as np

from libertem.udf import UDF

from utils import _mk_random


class PixelsumBaseUDF(UDF):
    def get_result_buffers(self):
        if self.meta.roi is not None:
            navsize = self.meta.roi.sum()
        else:
            navsize = np.prod(self.meta.dataset_shape.nav)
        return {
            'pixelsum_nav_raw': self.buffer(
                kind="single", dtype="float32", extra_shape=(navsize, )
            ),
            'pixelsum_sig_raw': self.buffer(
                kind="single", dtype="float32", extra_shape=self.meta.dataset_shape.sig
            ),
            'pixelsum_nav': self.buffer(
                kind="nav", dtype="float32"
            ),
            'pixelsum_sig': self.buffer(
                kind="sig", dtype="float32"
            ),
        }

    def get_tiling_preferences(self):
        # Example from SSB UDF: Target tile size depends on dataset shape
        # Make sure we have self.meta available at this stage
        result_size = np.prod(self.meta.dataset_shape.nav)
        target_size = 1024*1024
        return {
            "depth": max(1, target_size // result_size),
            "total_size": target_size,
        }

    def merge(self, dest, src):
        dest.pixelsum_nav_raw[:] += src.pixelsum_nav_raw
        dest.pixelsum_sig_raw[:] += src.pixelsum_sig_raw
        dest.pixelsum_nav[:] += src.pixelsum_nav
        dest.pixelsum_sig[:] += src.pixelsum_sig


class PixelsumPartitionUDF(PixelsumBaseUDF):
    def process_partition(self, partition):
        r = self.results
        r.pixelsum_nav[:] += np.sum(partition, axis=(-1, -2))
        r.pixelsum_sig[:] += np.sum(partition, axis=0)
        r.pixelsum_nav_raw[self.meta.slice.get(nav_only=True)] += np.sum(partition, axis=(-1, -2))
        r.pixelsum_sig_raw[self.meta.slice.get(sig_only=True)] += np.sum(partition, axis=0)


class PixelsumFrameUDF(PixelsumBaseUDF):
    def process_frame(self, frame):
        r = self.results
        r.pixelsum_nav[:] = np.sum(frame)
        r.pixelsum_sig[:] += frame
        r.pixelsum_nav_raw[self.meta.slice.get(nav_only=True)] = np.sum(frame)
        r.pixelsum_sig_raw[self.meta.slice.get(sig_only=True)] += frame


class PixelsumTileUDF(PixelsumBaseUDF):
    def process_tile(self, tile):
        r = self.results
        r.pixelsum_nav[:] += np.sum(tile, axis=(-1, -2))
        r.pixelsum_sig[:] += np.sum(tile, axis=0)
        r.pixelsum_nav_raw[self.meta.slice.get(nav_only=True)] += np.sum(tile, axis=(-1, -2))
        r.pixelsum_sig_raw[self.meta.slice.get(sig_only=True)] += np.sum(tile, axis=0)


def test_partition(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    dataset = lt_ctx.load(
        "memory",
        data=data,
        num_partitions=2,
        sig_dims=2
    )

    pixelsum = PixelsumPartitionUDF()
    res = lt_ctx.run_udf(dataset=dataset, udf=pixelsum)

    assert np.allclose(res['pixelsum_nav'].data, np.sum(data, axis=(2, 3)))
    assert np.allclose(res['pixelsum_nav_raw'].data.reshape((16, 16)), np.sum(data, axis=(2, 3)))
    assert np.allclose(res['pixelsum_sig'].data, np.sum(data, axis=(0, 1)))
    assert np.allclose(res['pixelsum_sig_raw'].data, np.sum(data, axis=(0, 1)))


def test_frame(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    dataset = lt_ctx.load(
        "memory",
        data=data,
        tileshape=(7, 16, 16),
        num_partitions=2,
        sig_dims=2
    )

    pixelsum = PixelsumFrameUDF()
    res = lt_ctx.run_udf(dataset=dataset, udf=pixelsum)

    assert np.allclose(res['pixelsum_nav'].data, np.sum(data, axis=(2, 3)))
    assert np.allclose(res['pixelsum_nav_raw'].data.reshape((16, 16)), np.sum(data, axis=(2, 3)))
    assert np.allclose(res['pixelsum_sig'].data, np.sum(data, axis=(0, 1)))
    assert np.allclose(res['pixelsum_sig_raw'].data, np.sum(data, axis=(0, 1)))


def test_tile(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    dataset = lt_ctx.load(
        "memory",
        data=data,
        tileshape=(7, 16, 16),
        num_partitions=2,
        sig_dims=2
    )

    pixelsum = PixelsumTileUDF()
    res = lt_ctx.run_udf(dataset=dataset, udf=pixelsum)

    assert np.allclose(res['pixelsum_nav'].data, np.sum(data, axis=(2, 3)))
    assert np.allclose(res['pixelsum_nav_raw'].data.reshape((16, 16)), np.sum(data, axis=(2, 3)))
    assert np.allclose(res['pixelsum_sig'].data, np.sum(data, axis=(0, 1)))
    assert np.allclose(res['pixelsum_sig_raw'].data, np.sum(data, axis=(0, 1)))


def test_roi(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    dataset = lt_ctx.load(
        "memory",
        data=data,
        tileshape=(7, 16, 16),
        num_partitions=2,
        sig_dims=2
    )

    roi = np.random.choice([True, False], dataset.shape.nav)

    pixelsum = PixelsumTileUDF()
    res = lt_ctx.run_udf(dataset=dataset, udf=pixelsum, roi=roi)

    assert np.allclose(res['pixelsum_nav'].raw_data, np.sum(data[roi], axis=(1, 2)))
    assert np.allclose(res['pixelsum_nav_raw'].data, np.sum(data[roi], axis=(1, 2)))
    assert np.allclose(res['pixelsum_sig'].raw_data, np.sum(data[roi], axis=0))
    assert np.allclose(res['pixelsum_sig_raw'].data, np.sum(data[roi], axis=0))



import numpy as np

import libertem.corrections.coordinates as c
from libertem.utils import rotate_deg


def test_scale():
    y, x = np.random.random((2, 7))
    factor = np.pi
    r_y, r_x = c.scale(factor) @ (y, x)
    assert np.allclose(r_y, y * factor)
    assert np.allclose(r_x, x * factor)


def test_rotate():
    y, x = np.random.random((2, 7))
    radians = np.pi/180*23
    r_y, r_x = c.identity() @ c.rotate(radians) @ (y, x)
    r_y2, r_x2 = rotate_deg(y, x, 23)
    assert np.allclose(r_y, r_y2)
    assert np.allclose(r_x, r_x2)


def test_rotate_deg():
    y, x = np.random.random((2, 7))
    degrees = 23
    r_y, r_x = c.rotate_deg(degrees) @ (y, x)
    r_y2, r_x2 = rotate_deg(y, x, degrees)
    assert np.allclose(r_y, r_y2)
    assert np.allclose(r_x, r_x2)


def test_flip_y():
    y, x = np.random.random((2, 7))
    r_y, r_x = c.flip_y() @ (y, x)
    assert np.allclose(r_y, -y)
    assert np.allclose(r_x, x)


def test_flip_x():
    y, x = np.random.random((2, 7))
    r_y, r_x = c.flip_x() @ (y, x)
    assert np.allclose(r_y, y)
    assert np.allclose(r_x, -x)



import pytest
import numpy as np

from libertem.udf.stddev import StdDevUDF
from libertem.udf.masks import ApplyMasksUDF


@pytest.mark.parametrize(
    'use_roi', [True, False]
)
def test_concurrent_executor(lt_ctx, concurrent_ctx, default_raw, use_roi):
    if use_roi:
        roi = np.random.choice([True, False], default_raw.shape.nav)
    else:
        roi = None
    mask = np.random.random(default_raw.shape.sig)

    def mask_factory():
        return mask

    load_params = {
        'filetype': 'raw',
        'path': default_raw._path,
        'nav_shape': default_raw.shape.nav,
        'sig_shape': default_raw.shape.sig,
        'dtype': default_raw.dtype
    }

    udfs = [StdDevUDF(), ApplyMasksUDF(mask_factories=[mask_factory])]
    ref_res = lt_ctx.run_udf(dataset=default_raw, udf=udfs, roi=roi)
    ds = concurrent_ctx.load(**load_params)
    res = concurrent_ctx.run_udf(dataset=ds, udf=udfs, roi=roi)

    assert len(ref_res) == len(res)

    for index, value in enumerate(ref_res):
        for key, ref in value.items():
            print("index", index, "key", key)
            assert np.allclose(ref.data, res[index][key].data, equal_nan=True)
        for key in res[index].keys():
            assert key in value



import os
import sys
import json

import numpy as np
import pytest
import cloudpickle

import libertem.api as lt

from libertem.analysis.raw import PickFrameAnalysis
from libertem.io.dataset.npy import NPYDatasetParams, NPYDataSet, read_npy_info
from libertem.io.dataset.base import (
    TilingScheme, BufferedBackend, MMapBackend, DirectBackend, DataSetException
)
from libertem.common import Shape
from libertem.common.math import prod
from libertem.common.buffers import reshaped_view
from libertem.udf.sumsigudf import SumSigUDF

from utils import dataset_correction_verification, ValidationUDF, roi_as_sparse


@pytest.mark.parametrize(
    "repeat", [*range(50)]
)
def test_read_npy_info(npy_random_array, repeat):
    npy_filepath, array = npy_random_array

    info = read_npy_info(npy_filepath)
    assert info.dtype == array.dtype
    assert info.shape == array.shape
    assert info.count == array.size
    # compute the header offset from the other direction !
    offset = os.stat(npy_filepath).st_size - array.nbytes
    assert info.offset == offset


def test_read_npy_info_fortran(npy_fortran_array):
    npy_filepath, array = npy_fortran_array
    assert array.flags.f_contiguous
    with pytest.raises(DataSetException):
        read_npy_info(npy_filepath)


def test_repr(default_npy):
    assert 'NPYDataSet' in repr(default_npy)


def test_get_fileset(default_npy):
    fileset = default_npy._get_fileset()
    assert len(fileset) == 1
    assert fileset[0]._path == default_npy._path


def test_detect_params(default_npy, default_raw_data, lt_ctx):
    params = NPYDataSet.detect_params(default_npy._path, lt_ctx.executor)
    assert (params['parameters']['nav_shape']
        + params['parameters']['sig_shape'])\
        == default_raw_data.shape


def test_detect_params_fail(lt_ctx):
    params = NPYDataSet.detect_params(555, lt_ctx.executor)
    assert not params


def test_apply_mask_analysis(default_npy, lt_ctx):
    mask = np.ones((128, 128))
    analysis = lt_ctx.create_mask_analysis(factories=[lambda: mask], dataset=default_npy)
    results = lt_ctx.run(analysis)
    assert results[0].raw_data.shape == (16, 16)


def test_sum_analysis(default_npy, lt_ctx):
    analysis = lt_ctx.create_sum_analysis(dataset=default_npy)
    results = lt_ctx.run(analysis)
    assert results[0].raw_data.shape == (128, 128)


def test_pick_analysis(default_npy, lt_ctx):
    analysis = PickFrameAnalysis(dataset=default_npy, parameters={"x": 15, "y": 15})
    results = lt_ctx.run(analysis)
    assert results[0].raw_data.shape == (128, 128)
    assert np.count_nonzero(results[0].raw_data) > 0


def test_comparison(default_npy, default_raw_data, lt_ctx_fast):
    udf = ValidationUDF(
        reference=reshaped_view(default_raw_data, (-1, *tuple(default_npy.shape.sig)))
    )
    lt_ctx_fast.run_udf(udf=udf, dataset=default_npy)


@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
def test_comparison_roi(default_npy, default_raw_data, lt_ctx_fast, as_sparse):
    roi = np.random.choice(
        [True, False],
        size=tuple(default_npy.shape.nav),
        p=[0.5, 0.5]
    )
    ref_data = default_raw_data[roi]
    if as_sparse:
        roi = roi_as_sparse(roi)
    udf = ValidationUDF(reference=ref_data)
    lt_ctx_fast.run_udf(udf=udf, dataset=default_npy, roi=roi)


def test_pickle_is_small(default_npy):
    pickled = cloudpickle.dumps(default_npy)
    cloudpickle.loads(pickled)

    # let's keep the pickled dataset size small-ish:
    assert len(pickled) < 2 * 1024


def test_cache_key_json_serializable(default_npy):
    json.dumps(default_npy.get_cache_key())


def test_message_converter_direct():
    src = {
        "type": "NPY",
        "path": "p",
        "nav_shape": [16, 16],
        "sig_shape": [8, 8],
    }
    converted = NPYDatasetParams().convert_to_python(src)
    assert converted == {
        "path": "p",
        "nav_shape": (16, 16),
        "sig_shape": (8, 8),
    }


def test_bad_sig_dims(lt_ctx):
    with pytest.raises(DataSetException):
        lt_ctx.load(
            "npy",
            path='anything.npy',
            sig_shape=(1, 2, 3),
            sig_dims=1,
        )


def test_bad_sig_dims2(lt_ctx):
    with pytest.raises(DataSetException):
        lt_ctx.load(
            "npy",
            path='anything.npy',
            sig_dims=None,
        )


def test_auto_sig_dims(lt_ctx):
    ds = NPYDataSet(path='anything.npy',
                    sig_shape=(1, 2, 3),
                    sig_dims=None)
    assert ds._sig_dims == 3


def test_shape_arg_smallernav(default_npy, default_npy_filepath, default_raw_data, lt_ctx):
    nav_shape, sig_shape = default_raw_data.shape[:2], default_raw_data.shape[2:]
    smaller_nav_shape = tuple(n - 1 for n in nav_shape)
    ds = lt_ctx.load(
            "npy",
            path=default_npy_filepath,
            sig_dims=2,
            nav_shape=smaller_nav_shape,
        )
    assert tuple(ds.shape.sig) == sig_shape
    assert tuple(ds.shape.nav) == smaller_nav_shape
    lt_ctx: lt.Context
    result = lt_ctx.map(ds, lambda x: True)
    assert result.data.shape == smaller_nav_shape


def test_shape_arg_flatnav(default_npy, default_npy_filepath, default_raw_data, lt_ctx):
    nav_shape, sig_shape = default_raw_data.shape[:2], default_raw_data.shape[2:]
    flat_nav_shape = (prod(nav_shape),)
    ds = lt_ctx.load(
            "npy",
            path=default_npy_filepath,
            sig_dims=2,
            nav_shape=flat_nav_shape,
        )
    assert tuple(ds.shape.sig) == sig_shape
    assert tuple(ds.shape.nav) == flat_nav_shape
    lt_ctx: lt.Context
    result = lt_ctx.map(ds, lambda x: True)
    assert result.data.shape == flat_nav_shape


@pytest.mark.with_numba
@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
def test_roi_1(default_npy, lt_ctx, as_sparse):
    p = next(default_npy.get_partitions())
    roi = np.zeros(p.meta.shape.flatten_nav().nav, dtype=bool)
    roi[0] = 1
    tiles = []
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=Shape((1, 128, 128), sig_dims=2),
        dataset_shape=default_npy.shape,
    )
    if as_sparse:
        roi = roi_as_sparse(roi)
    for tile in p.get_tiles(tiling_scheme=tiling_scheme, dest_dtype="float32", roi=roi):
        print("tile:", tile)
        tiles.append(tile)
    assert len(tiles) == 1
    assert tiles[0].tile_slice.origin == (0, 0, 0)
    assert tuple(tiles[0].tile_slice.shape) == (1, 128, 128)


@pytest.mark.with_numba
@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
def test_roi_2(default_npy, lt_ctx, as_sparse):
    p = next(default_npy.get_partitions())
    roi = np.zeros(p.meta.shape.flatten_nav(), dtype=bool)
    stackheight = 4
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=Shape((stackheight, 128, 128), sig_dims=2),
        dataset_shape=default_npy.shape,
    )
    roi[0:stackheight + 2] = 1
    if as_sparse:
        roi = roi_as_sparse(roi)
    tiles = p.get_tiles(tiling_scheme=tiling_scheme, dest_dtype="float32", roi=roi)
    tiles = list(tiles)


@pytest.mark.parametrize(
    "io_backend", (
        BufferedBackend(),
        MMapBackend(),
    ),
)
def test_positive_sync_offset(lt_ctx, npy_8x8x8x8_ds, npy_8x8x8x8_path, io_backend):
    udf = SumSigUDF()
    sync_offset = 2

    ds_with_offset = NPYDataSet(
        path=npy_8x8x8x8_path,
        sync_offset=sync_offset,
        io_backend=io_backend,
    )
    ds_with_offset.set_num_cores(4)
    ds_with_offset = ds_with_offset.initialize(lt_ctx.executor)
    ds_with_offset.check_valid()

    p0 = next(ds_with_offset.get_partitions())
    assert p0._start_frame == 2
    assert p0.slice.origin == (0, 0, 0)

    tileshape = Shape(
        (4,) + tuple(ds_with_offset.shape.sig),
        sig_dims=ds_with_offset.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds_with_offset.shape,
    )

    tiles = p0.get_tiles(tiling_scheme)
    t0 = next(tiles)
    assert tuple(t0.tile_slice.origin) == (0, 0, 0)
    for _ in tiles:
        pass

    for p in ds_with_offset.get_partitions():
        for t in p.get_tiles(tiling_scheme=tiling_scheme):
            pass

    assert p.slice.origin == (48, 0, 0)
    assert p.slice.shape[0] == 16

    result = lt_ctx.run_udf(dataset=npy_8x8x8x8_ds, udf=udf)
    result = result['intensity'].raw_data[sync_offset:]

    result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf)
    result_with_offset = result_with_offset['intensity'].raw_data[
        :ds_with_offset._meta.image_count - sync_offset
    ]

    assert np.allclose(result, result_with_offset)


@pytest.mark.parametrize(
    "io_backend", (
        BufferedBackend(),
        MMapBackend(),
    ),
)
def test_negative_sync_offset(lt_ctx, npy_8x8x8x8_ds, npy_8x8x8x8_path, io_backend):
    udf = SumSigUDF()
    sync_offset = -2

    ds_with_offset = NPYDataSet(
        path=npy_8x8x8x8_path,
        sync_offset=sync_offset,
        io_backend=io_backend,
    )
    ds_with_offset.set_num_cores(4)
    ds_with_offset = ds_with_offset.initialize(lt_ctx.executor)
    ds_with_offset.check_valid()

    p0 = next(ds_with_offset.get_partitions())
    assert p0._start_frame == -2
    assert p0.slice.origin == (0, 0, 0)

    tileshape = Shape(
        (4,) + tuple(ds_with_offset.shape.sig),
        sig_dims=ds_with_offset.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds_with_offset.shape,
    )

    tiles = p0.get_tiles(tiling_scheme)
    t0 = next(tiles)
    assert tuple(t0.tile_slice.origin) == (2, 0, 0)
    for _ in tiles:
        pass

    for p in ds_with_offset.get_partitions():
        for t in p.get_tiles(tiling_scheme=tiling_scheme):
            pass

    assert p.slice.origin == (48, 0, 0)
    assert p.slice.shape[0] == 16

    result = lt_ctx.run_udf(dataset=npy_8x8x8x8_ds, udf=udf)
    result = result['intensity'].raw_data[:npy_8x8x8x8_ds._meta.image_count - abs(sync_offset)]

    result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf)
    result_with_offset = result_with_offset['intensity'].raw_data[abs(sync_offset):]

    assert np.allclose(result, result_with_offset)


@pytest.mark.skipif(
    sys.platform.startswith("darwin"),
    reason="No support for direct I/O on Mac OS X"
)
def test_load_direct(lt_ctx, npy_8x8x8x8_path):
    ds_direct = lt_ctx.load(
        "npy",
        path=npy_8x8x8x8_path,
        io_backend=DirectBackend(),
    )
    analysis = lt_ctx.create_sum_analysis(dataset=ds_direct)
    lt_ctx.run(analysis)


@pytest.mark.parametrize(
    "io_backend", (
        BufferedBackend(),
        MMapBackend(),
    ),
)
def test_reshape_nav(lt_ctx, npy_8x8x8x8_ds, npy_8x8x8x8_path, io_backend):
    udf = SumSigUDF()

    ds_with_1d_nav = lt_ctx.load(
        "npy",
        path=npy_8x8x8x8_path,
        nav_shape=(64,),
        io_backend=io_backend,
    )
    result_with_1d_nav = lt_ctx.run_udf(dataset=ds_with_1d_nav, udf=udf)
    result_with_1d_nav = result_with_1d_nav['intensity'].raw_data

    result_with_2d_nav = lt_ctx.run_udf(dataset=npy_8x8x8x8_ds, udf=udf)
    result_with_2d_nav = result_with_2d_nav['intensity'].raw_data

    ds_with_3d_nav = lt_ctx.load(
        "npy",
        path=npy_8x8x8x8_path,
        nav_shape=(2, 4, 8),
    )
    result_with_3d_nav = lt_ctx.run_udf(dataset=ds_with_3d_nav, udf=udf)
    result_with_3d_nav = result_with_3d_nav['intensity'].raw_data

    assert np.allclose(result_with_1d_nav, result_with_2d_nav, result_with_3d_nav)


def test_different_sig_shape(lt_ctx, npy_8x8x8x8_path):
    sig_shape = (4, 4)

    ds = lt_ctx.load(
        "npy",
        path=npy_8x8x8x8_path,
        sig_shape=sig_shape,
    )

    assert ds._meta.image_count == 256


@pytest.mark.parametrize(
    "io_backend", (
        BufferedBackend(),
        MMapBackend(),
    ),
)
def test_extra_data_at_the_end(lt_ctx, npy_8x8x8x8_path, io_backend):
    """
    If there is extra data at the end of the file, make sure it is cut off
    """
    sig_shape = (3, 3)

    ds = lt_ctx.load(
        "npy",
        path=npy_8x8x8x8_path,
        sig_shape=sig_shape,
        io_backend=io_backend,
    )

    assert ds._meta.image_count == 455


def test_offset_greater_than_image_count(lt_ctx, npy_8x8x8x8_path):
    sync_offset = 70

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "npy",
            path=npy_8x8x8x8_path,
            sync_offset=sync_offset,
        )
    assert e.match(
        r"offset should be in \(-64, 64\), which is \(-image_count, image_count\)"
    )


@pytest.mark.parametrize(
    "with_roi", (True, False)
)
def test_correction_default(default_npy, lt_ctx, with_roi):
    ds = default_npy

    if with_roi:
        roi = np.zeros(ds.shape.nav, dtype=bool)
        roi[:1] = True
    else:
        roi = None

    dataset_correction_verification(ds=ds, roi=roi, lt_ctx=lt_ctx)


def test_diagnostics(default_npy):
    assert {"name": "dtype", "value": "float32"} in default_npy.get_diagnostics()


def test_bad_params(ds_params_tester, standard_bad_ds_params, default_npy):
    args = ("npy", default_npy._path)
    for params in standard_bad_ds_params:
        ds_params_tester(*args, **params)



import os
from unittest import mock

import pytest

from libertem.io.fs import get_fs_listing


def test_get_fs_listing_permission_of_subdir(tmpdir_factory):
    tmpdir = tmpdir_factory.mktemp('get_fs_listing')
    tmpdir.mkdir('sub1')
    no_access_dir = tmpdir.mkdir('sub2')

    orig_stat = os.stat

    def mock_stat(path, *args, **kwargs):
        if os.path.normpath(path) == os.path.normpath(no_access_dir):
            raise PermissionError(f"[Errno 13] Access is denied: '{path}'")
        return orig_stat(path, *args, **kwargs)

    # patch os.stat to fail for `sub2`:
    with mock.patch('os.stat', side_effect=mock_stat):
        listing = get_fs_listing(tmpdir)

        dir_names = [
            entry['name']
            for entry in listing['dirs']
        ]
        assert 'sub2' not in dir_names

        # no direct access to `sub2`
        with pytest.raises(PermissionError):
            os.stat(no_access_dir)



from typing import Union
from collections.abc import Iterable

import numpy as np

from .sparse import assert_sparse


_prod_accepted = (
    int, bool,
    np.bool_, np.signedinteger, np.unsignedinteger
)

ProdAccepted = Union[
    int, bool,
    np.bool_, np.signedinteger, np.unsignedinteger
]


def prod(iterable: Iterable[ProdAccepted]):
    '''
    Safe product for large integer size calculations.

    :meth:`numpy.prod` uses 32 bit for default :code:`int` on Windows 64 bit. This
    function uses infinite width integers to calculate the product and
    throws a ValueError if it encounters types other than the supported ones.
    '''
    result = 1

    for item in iterable:
        if isinstance(item, _prod_accepted):
            result *= int(item)
        else:
            raise ValueError()
    return result


def count_nonzero(array) -> int:
    nnz: int  # for mypy
    try:
        nnz = np.count_nonzero(array)
    except TypeError:
        sparse_a = assert_sparse(array)
        nnz = sparse_a.nnz
    return nnz


def flat_nonzero(array):
    return array.flatten().nonzero()[0]


def make_2D_square(shape: tuple[int, ...]) -> tuple[int, ...]:
    """
    Convert the 1D shape tuple into a square 2D shape tuple
    if the size of shape is a square number

    Non-square dim and len(shape) != 1 are passed directly through.
    shape (1,) is considered square and returns (1, 1)

    Parameters
    ----------
    shape : Tuple[int, ...]
        1D shape tuple[int]

    Returns
    -------
    Tuple[int, ...]
        2D shape tuple[int, int] if shape is 1D and contains a square
        number of elements, else the input shape is returned
    """
    if len(shape) != 1:
        return shape
    size = prod(shape)
    if size < 1:
        raise ValueError('Zero or negative shape.size')
    dim, remainder = divmod(np.sqrt(size), 1)
    if remainder == 0:
        return (dim,) * 2
    return shape


def _sparse_ndenumerate(array):
    flat_array = array.flatten()
    nonzero = flat_nonzero(flat_array)
    for idx in nonzero:
        coords = np.unravel_index(idx, array.shape)
        yield coords, flat_array[idx]


def ndenumerate(array):
    try:
        yield from np.ndenumerate(array)
    except RuntimeError:
        yield from _sparse_ndenumerate(array)



from libertem.io.utils import get_partition_shape
from libertem.io.dataset.base.utils import FileTree
from libertem.common import Shape

from utils import MockFile


def test_partition_shape_1d():
    ds_shape = Shape((15, 16, 16), sig_dims=2)
    pshape = get_partition_shape(
        dataset_shape=ds_shape,
        target_size_items=256*1024,
        min_num=2
    )
    assert pshape == (7,)


def test_partition_shape_1():
    assert get_partition_shape(Shape((15, 16, 16), sig_dims=2), target_size_items=512) == (
        (2,)
    )


def test_partition_shape_2():
    assert get_partition_shape(Shape((1, 15, 16, 16), sig_dims=2), target_size_items=512) == (
        (1, 2,)
    )


def test_partition_shape_3():
    assert get_partition_shape(Shape((15, 15, 16, 16), sig_dims=2), target_size_items=15*512) == (
        (2, 15,)
    )


def test_partition_shape_4():
    assert get_partition_shape(
        Shape((128, 15, 15, 16, 16), sig_dims=2),
        target_size_items=15*512
    ) == (
        (1, 2, 15,)
    )


def test_partition_shape_5():
    assert get_partition_shape(
        Shape((2, 16, 16), sig_dims=2),
        target_size_items=512,
        min_num=3
    ) == (
        (1,)
    )


def test_partition_shape_small():
    assert get_partition_shape(Shape((15, 16, 16), sig_dims=2), target_size_items=4) == (
        (1,)
    )


def mock_files(num):
    return [
        MockFile(start_idx=x, end_idx=x + 8)
        for x in range(0, num*8, 8)
    ]


def test_make_file_tree_1():
    files = mock_files(1)
    tree = FileTree.make(files)
    print(tree)

    files = mock_files(2)
    tree = FileTree.make(files)
    print(tree)

    files = mock_files(3)
    tree = FileTree.make(files)
    print(tree)

    files = mock_files(4)
    tree = FileTree.make(files)
    print(tree)

    files = mock_files(5)
    tree = FileTree.make(files)
    print(tree)


def test_search():
    files = mock_files(5)
    tree = FileTree.make(files)
    print(tree)

    for i in range(8):
        assert tree.search_start(i)[1].start_idx == 0

    for i in range(16, 24):
        assert tree.search_start(i)[1].start_idx == 16

    for i in range(24, 32):
        assert tree.search_start(i)[1].start_idx == 24

    for i in range(32, 40):
        assert tree.search_start(i)[1].start_idx == 32


def test_ft_single_frame_single_file():
    f1 = MockFile(start_idx=0, end_idx=1)
    files = [f1]

    ft = FileTree.make(files)

    assert ft.search_start(0) == (0, f1)

    print(ft, ft.__dict__)


def test_ft_single_frame_multiple_files():
    f1 = MockFile(start_idx=0, end_idx=1)
    f2 = MockFile(start_idx=1, end_idx=2)
    f3 = MockFile(start_idx=2, end_idx=3)
    f4 = MockFile(start_idx=3, end_idx=4)
    files = [f1, f2, f3, f4]

    ft = FileTree.make(files)

    assert ft.search_start(0) == (0, f1)

    print(ft, ft.__dict__)


def test_ft_multi_frame_multiple_files_even():
    f1 = MockFile(start_idx=0, end_idx=2)
    f2 = MockFile(start_idx=2, end_idx=4)
    f3 = MockFile(start_idx=4, end_idx=6)
    f4 = MockFile(start_idx=6, end_idx=8)
    files = [f1, f2, f3, f4]

    ft = FileTree.make(files)
    print(ft)

    assert ft.search_start(0) == (0, f1)
    assert ft.search_start(2) == (1, f2)
    assert ft.search_start(4) == (2, f3)
    assert ft.search_start(6) == (3, f4)


def test_ft_multi_frame_multiple_files_odd():
    f1 = MockFile(start_idx=0, end_idx=1)
    f2 = MockFile(start_idx=2, end_idx=3)
    f3 = MockFile(start_idx=4, end_idx=5)
    f4 = MockFile(start_idx=6, end_idx=7)
    f5 = MockFile(start_idx=8, end_idx=9)
    files = [f1, f2, f3, f4, f5]

    ft = FileTree.make(files)
    print(ft)

    assert ft.search_start(0) == (0, f1)
    assert ft.search_start(2) == (1, f2)
    assert ft.search_start(4) == (2, f3)
    assert ft.search_start(6) == (3, f4)
    assert ft.search_start(8) == (4, f5)



from functools import partial
from typing import Any, Optional
from collections.abc import Iterable
import contextlib
from collections import defaultdict, OrderedDict

import numpy as np
import dask
from dask import delayed
import dask.array as da

from libertem.io.corrections import CorrectionSet
from libertem.io.dataset.base import DataSet
from libertem.utils.devices import detect

from .base import BaseJobExecutor
from libertem.common.executor import Environment, TaskCommHandler, TaskProtocol
from libertem.common.scheduler import Worker, WorkerSet

from ..common.buffers import BufferWrapper
from ..common.math import prod
from ..udf.base import (
    UDFMergeAllMixin, UDFRunner, UDF, UDFData, MergeAttrMapping,
    get_resources_for_backends, UDFResults, BackendSpec
)

from .utils.dask_buffer import DaskBufferWrapper, DaskPreallocBufferWrapper, DaskResultBufferWrapper
from .utils import delayed_unpack


class DelayedUDFRunner(UDFRunner):
    def __init__(self, udfs: list[UDF], debug: bool = False, progress_reporter: Any = None):
        self._part_results = defaultdict(dict)
        super().__init__(udfs, debug=debug)

    @staticmethod
    def _make_udf_result(udfs: Iterable[UDF], damage: BufferWrapper) -> "UDFResults":
        udf_results = UDFRunner._make_udf_result(udfs, damage)
        buffers = udf_results.buffers
        damage = udf_results.damage
        new_buffers = tuple(
            {
                k: DaskResultBufferWrapper.from_buffer_wrapper(v)
                for k, v in bufs.items()
            }
            for bufs in buffers
        )
        return UDFResults(
            buffers=new_buffers,
            damage=damage,
        )

    def _apply_part_result(self, udfs: Iterable[UDF], damage, part_results, task):
        for part_results_udf, udf in zip(part_results, udfs):
            # Allow user to define an alternative merge strategy
            # using dask-compatible functions. In the Delayed case we
            # won't be getting partial results with damage anyway.
            # Currently there is no interface to provide all of the results
            # to udf.merge at once and in the correct order, so I am accumulating
            # results in self._part_results[udf] = {partition_slice_roi: part_results, ...}
            if (isinstance(udf, UDFMergeAllMixin)
                    or (type(udf).merge is UDF.merge and not udf.requires_custom_merge_all)):
                if self._accumulate_part_results(udf, part_results_udf, task):
                    try:
                        parts = {
                            key: value.get_proxy()
                            for key, value in self._part_results[udf].items()
                        }
                        udf._do_merge_all(parts)
                    finally:
                        self._part_results.pop(udf, None)
                continue

            # In principle we can requrire that sig merges are done sequentially
            # by using dask.graph_manipulation.bind, this could improve thread
            # safety on the sig buffer. In the current implem it's not necessary as
            # we are calling delayed_apply_part_result sequentially on results
            # and replacing all buffers with their partially merged versions
            # on each call, so we are implicitly sequential when updating
            # try:
            #     bind_to = udf.prior_merge
            # except AttributeError:
            #     bind_to = None

            structure = structure_from_task([udf], task)[0]
            flat_structure = delayed_unpack.flatten_nested(structure)
            flat_mapping = delayed_unpack.build_mapping(structure)

            src = part_results_udf.get_proxy()
            src_dict = {k: b for k, b in src._dict.items()}

            dest_dict = {}
            for k, b_decl in udf.get_result_buffers().items():
                # We skip result_only buffers since they are not expected
                # in the dest_dict by the default merge function
                if b_decl.use == 'result_only':
                    continue
                b = udf.results.get_buffer(k)
                view = b.get_view_for_partition(task.partition)
                # Handle result-only buffers
                if view is not None:
                    try:
                        dest_dict[k] = view.unwrap_sliced()
                    except AttributeError:
                        # Handle kind='single' buffers
                        # Could handle sig buffers this way too if we assert no chunking
                        dest_dict[k] = view

            # Run the udf.merge function in a wrapper which flattens
            # the results to a flat list which can be unpacked from the Delayed
            # This step makes a copy of the buffer view inside the delayed
            # call because this default Dask behaviour is to give a read-only
            # view of the arguments to the delayed function
            merged_del = delayed(merge_wrap, nout=len(flat_mapping))
            # need to copy the damage, as it will be overwritten in the next loop iteration,
            # which is independent from actually evaluating the delayed object:
            merged_del_result = merged_del(udf, dest_dict, src_dict, damage.raw_data.copy())

            # The following is needed when binding sequential updates on sig buffers
            # if bind_to is not None:
            #     merged_del = delayed_bind([*merged_del_result],
            #                               [*bind_to])
            # if any([b.kind != 'nav' for _, b in udf.results.items()]):
            #     # have sig result, gotta do a bind merge next time round
            #     udf.prior_merge = merged_del_result

            # We now build the result back into dask arrays and assign them
            # into the appropriate slices of the full result buffers
            wrapped_res = delayed_to_buffer_wrappers(merged_del_result, flat_structure,
                                                     task.partition, as_buffer=False)
            renested = delayed_unpack.rebuild_nested(wrapped_res, flat_mapping)

            # Assign into the result buffers with the partially merged version
            # This is OK because we are calling this part merge function sequentially
            # so each new call gets the most recent version of the buffer in the
            # dask task graph
            udf.set_views_for_partition(task.partition)
            dest = udf.results.get_proxy()
            merged = MergeAttrMapping(renested)
            for k in dest:
                getattr(dest, k)[:] = getattr(merged, k)

        v = damage.get_view_for_partition(task.partition)
        v[:] = True

    def _accumulate_part_results(self, udf, part_results, task):
        """
        If the udf has a merge_all method, this function is used
        to accumulate dask array-backed partial results on the executor
        so that the merge_all can be called on all of them at once

        Ensures that the results are correctly ordered and complete
        before allowing merge_all to be called
        """
        buf = next(iter(part_results.values()))  # get the first buffer
        slice_with_roi = buf._slice_for_partition(task.partition)
        self._part_results[udf][slice_with_roi] = part_results

        # number of frames in dataset
        if udf.meta.roi is not None:
            target_coverage = np.count_nonzero(udf.meta.roi)
        else:
            target_coverage = prod(task.partition.meta.shape.nav)
        # number of frames we have results for
        current_coverage = sum(prod(k.shape.nav) for k in self._part_results[udf].keys())
        if target_coverage == current_coverage:
            ordered_results = sorted(self._part_results[udf].items(),
                                     key=lambda kv: kv[0].origin[0])
            self._part_results[udf] = OrderedDict(ordered_results)
            return True
        elif current_coverage > target_coverage:
            raise RuntimeError('More frames accumulated than ROI specifies - '
                              f'target {target_coverage} - processed {current_coverage}')
        return False

    def results_for_dataset_sync(self, dataset: DataSet, executor: 'DelayedJobExecutor',
            roi: Optional[np.ndarray] = None, progress: bool = False,
            corrections: Optional[CorrectionSet] = None, backends: Optional[BackendSpec] = None,
            dry: bool = False) -> Iterable[tuple]:

        executor.register_master_udfs(self._udfs)

        return super().results_for_dataset_sync(
            dataset, executor, roi=roi, progress=progress,
            corrections=corrections, backends=backends, dry=dry
        )


class DelayedJobExecutor(BaseJobExecutor):
    """
    :class:`~libertem.common.executor.JobExecutor` that uses dask.delayed to execute tasks.

    .. versionadded:: 0.9.0

    Highly experimental at this time!
    """
    def __init__(self):
        # Only import if actually instantiated, i.e. will likely be used
        import libertem.preload  # noqa: 401
        self._udfs = None

    @contextlib.contextmanager
    def scatter(self, obj):
        yield delayed(obj)

    def cancel(self, cancel_id: Any):
        pass

    def run_tasks(
        self,
        tasks: Iterable[TaskProtocol],
        params_handle: Any,
        cancel_id: Any,
        task_comm_handler: TaskCommHandler,
    ):
        """
        Wraps the call task() such that it returns a flat list
        of results, then unpacks the Delayed return value into
        the normal

            :code:`tuple(udf.results for udf in self._udfs)`

        where the buffers inside udf.results are dask arrays instead
        of normal np.arrays

        Needs a reference to the udfs on the master node so
        that the results structure can be inferred. This reference is
        found in self._udfs, which is set with the method:

            :code:`executor.register_master_udfs(udfs)`

        called from :meth:`DelayedUDFRunner.results_for_dataset_sync`
        """
        env = Environment(threads_per_worker=1, threaded_executor=True)
        for task in tasks:
            structure = structure_from_task(self._udfs, task)
            flat_structure = delayed_unpack.flatten_nested(structure)
            flat_mapping = delayed_unpack.build_mapping(structure)
            flat_result_task = partial(task_wrap, task)
            result = delayed(flat_result_task, nout=len(flat_structure))(env=env,
                                                                         params=params_handle)
            wrapped_res = delayed_to_buffer_wrappers(result, flat_structure, task.partition,
                                                     roi=self._udfs[0].meta.roi)
            renested = delayed_unpack.rebuild_nested(wrapped_res, flat_mapping)
            result = tuple(UDFData(data=res) for res in renested)
            yield result, task

    def run_function(self, fn, *args, **kwargs):
        result = fn(*args, **kwargs)
        return result

    def run_delayed(self, fn, *args, _delayed_kwargs=None, **kwargs):
        if _delayed_kwargs is None:
            _delayed_kwargs = {}
        result = delayed(fn, **_delayed_kwargs)(*args, **kwargs)
        return result

    def map(self, fn, iterable):
        return [fn(item)
                for item in iterable]

    def run_each_host(self, fn, *args, **kwargs):
        return {"localhost": fn(*args, **kwargs)}

    def run_each_worker(self, fn, *args, **kwargs):
        return {"delayed": fn(*args, **kwargs)}

    def get_available_workers(self):
        resources = {"compute": 1, "CPU": 1}
        # We don't know at this time,
        # but assume one worker per CPU
        devices = detect()
        return WorkerSet([
            Worker(
                name='delayed', host='localhost',
                resources=resources,
                nthreads=len(devices['cpus']),
            )
        ])

    def modify_buffer_type(self, buf):
        """
        Convert existing buffers from BufferWrapper to DaskBufferWrapper

        A refactoring of the UDF backend would remove the need for this method.

        :meta private:
        """
        return DaskBufferWrapper.from_buffer(buf)

    def register_master_udfs(self, udfs):
        """
        Give the executor a reference to the udfs instantiated
        on the main node, for introspection purposes

        :meta private:
        """
        self._udfs = udfs

    def _compute(self, *args, udfs=None, user_backends=None, traverse=True, **kwargs):
        """
        Acts as dask.compute(*args, **kwargs) but with knowledge
        of Libertem data structures and compute resources
        """
        if 'resources' in kwargs:
            if udfs is not None:
                raise ValueError('Cannot specify both udfs for resources and resources to use')
            resources = kwargs.get('resources')
        elif udfs is not None:
            resources = self.get_resources(udfs, user_backends=user_backends)
        else:
            resources = None
        kwargs['resources'] = resources

        to_unpack = tuple(a for a in args)
        unwrapped_args = tuple(self.unwrap_results(a) for a in to_unpack)
        results = dask.compute(*unwrapped_args, traverse=traverse, **kwargs)
        if len(args) == 1:
            if len(results) > 1:
                raise RuntimeWarning(f'Unexpected number of results {len(results)} '
                                     'from dask.compute, dropping extras')
            results = results[0]
        return results

    @staticmethod
    def get_resources_from_udfs(udfs, user_backends=None):
        """
        Returns the resources required by the udfs passed as
        argument, excluding those not in the tuple user_backends
        """
        if user_backends is None:
            user_backends = tuple()
        if isinstance(udfs, UDF):
            udfs = [udfs]
        backends = [udf.get_backends() for udf in udfs]
        return get_resources_for_backends(backends, user_backends)

    @staticmethod
    def unwrap_results(results):
        unpackable = {**delayed_unpack.default_unpackable(),
                      UDFData: lambda x: x._data.items(),
                      DaskPreallocBufferWrapper: lambda x: [(0, x.data)],
                      DaskBufferWrapper: lambda x: [(0, x.data)],
                      }

        res_unpack = delayed_unpack.flatten_nested(results, unpackable_types=unpackable)
        flat_mapping = delayed_unpack.build_mapping(results, unpackable_types=unpackable)
        flat_mapping_reduced = [el[:-1] if issubclass(el[-1][0], BufferWrapper) else el
                                for el in flat_mapping]
        return delayed_unpack.rebuild_nested(res_unpack, flat_mapping_reduced)

    def get_udf_runner(self) -> type['UDFRunner']:
        return DelayedUDFRunner


def make_copy(array_dict):
    for k, v in array_dict.items():
        if not v.flags['WRITEABLE']:
            array_dict[k] = v.copy()
    return array_dict


def merge_wrap(udf, dest_dict, src_dict, raw_damage):
    """
    The function called as delayed, acting as a wrapper
    to return a flat list of results rather than a structure
    of UDFData or MergeAttrMapping

    :meta private:
    """
    # Have to make a copy of dest buffers because Dask brings
    # data into the delayed function as read-only np arrays
    # I experimented with setting WRITEABLE to True but this
    # resulted in errors in the final array
    dest_dict = make_copy(dest_dict)

    dest = MergeAttrMapping(dest_dict)
    src = MergeAttrMapping(src_dict)

    # In place merge into the copy of dest
    udf.meta.set_valid_nav_mask(raw_damage)
    udf.merge(dest=dest, src=src)
    # Return flat list of results so they can be unpacked later
    return delayed_unpack.flatten_nested(dest._dict)


def task_wrap(task, *args, **kwargs):
    """
    Flatten the structure tuple(udf.results for udf in self._udfs)
    where udf.results is an instance of UDFData(data={'name':BufferWrapper,...})
    into a simple list [np.ndarray, np.ndarray, ...]

    :meta private:
    """
    res = task(*args, **kwargs)
    res = tuple(r._data for r in res)
    flat_res = delayed_unpack.flatten_nested(res)
    return [r._data for r in flat_res]


def structure_from_task(udfs, task):
    """
    Based on the instantiated whole dataset UDFs and the task
    information, build a description of the expected UDF results
    for the task's partition like:

    :code:`({'buffer_name': StructDescriptor(shape, dtype, extra_shape, buffer_kind), ...}, ...)`

    :meta private:
    """
    structure = []
    for udf in udfs:
        res_data = {}
        for buffer_name, buffer in udf.results.items():
            part_buf_extra_shape = buffer.extra_shape
            buffer.set_shape_partition(task.partition, roi=buffer._roi)
            part_buf_shape = buffer.shape
            part_buf_dtype = buffer.dtype
            res_data[buffer_name] = \
                delayed_unpack.StructDescriptor(np.ndarray,
                                                shape=part_buf_shape,
                                                dtype=part_buf_dtype,
                                                kind=buffer.kind,
                                                extra_shape=part_buf_extra_shape)
        results_container = res_data
        structure.append(results_container)
    return tuple(structure)


def delayed_to_buffer_wrappers(flat_delayed, flat_structure, partition, roi=None, as_buffer=True):
    """
    Take the iterable Delayed results object, and re-wrap each Delayed object
    back into a BufferWrapper wrapping a dask.array of the correct shape and dtype

    :meta private:
    """
    wrapped_res = []
    for el, descriptor in zip(flat_delayed, flat_structure):
        buffer_kind = descriptor.kwargs.pop('kind')
        extra_shape = descriptor.kwargs.pop('extra_shape')
        buffer_dask = da.from_delayed(el, *descriptor.args, **descriptor.kwargs)
        if as_buffer:
            buffer = DaskBufferWrapper(buffer_kind,
                                       extra_shape=extra_shape,
                                       dtype=descriptor.kwargs['dtype'])
            # Need to test whether roi=None here is a problem
            buffer.set_shape_partition(partition, roi=roi)
            buffer.replace_array(buffer_dask)
            wrapped_res.append(buffer)
        else:
            wrapped_res.append(buffer_dask)
    return wrapped_res



import numpy as np

from libertem.udf import UDF


class SumSigUDF(UDF):
    """
    Sum over the signal axes. For each navigation position, the sum of all pixels is calculated.

    Examples
    --------
    >>> udf = SumSigUDF()
    >>> result = ctx.run_udf(dataset=dataset, udf=udf)
    >>> np.array(result["intensity"]).shape
    (16, 16)
    """

    def get_backends(self):
        return self.BACKEND_ALL

    def get_result_buffers(self):
        ""
        dtype = np.result_type(self.meta.input_dtype, np.float32)
        return {
            'intensity': self.buffer(
                kind="nav", dtype=dtype, where='device'
            ),
        }

    def process_tile(self, tile):
        ""
        self.results.intensity[:] += self.forbuf(
            np.sum(
                # Flatten and sum axis 1 for cupyx.scipy.sparse support
                tile.reshape((tile.shape[0], -1)),
                axis=1
            ),
            self.results.intensity
        )


def run_sumsig(ctx, dataset):
    udf = SumSigUDF()
    pass_results = ctx.run_udf(dataset=dataset, udf=udf)
    return pass_results
